@@ -12,6 +12,7 @@
 #include <pwd.h>
 #include <sys/param.h>
 #include <libzscanner/scanner.h>
+#include <sys/un.h>
 
 #include <lua.h>
 #include <lualib.h>
@@ -21,6 +22,7 @@
 #include "kresconfig.h"
 #include "daemon/engine.h"
 #include "daemon/ffimodule.h"
+#include "daemon/worker.h"
 #include "lib/nsrep.h"
 #include "lib/cache/api.h"
 #include "lib/defines.h"
@@ -35,7 +37,7 @@
 #define LRU_REP_SIZE (LRU_RTT_SIZE / 4) /**< NS reputation cache size */
 #endif
 #ifndef LRU_COOKIES_SIZE
-	#ifdef ENABLE_COOKIES
+	#if ENABLE_COOKIES
 	#define LRU_COOKIES_SIZE LRU_RTT_SIZE /**< DNS cookies cache size. */
 	#else
 	#define LRU_COOKIES_SIZE LRU_ASSOC /* simpler than guards everywhere */
@@ -44,7 +46,7 @@
 
 /**@internal Maximum number of incomplete TCP connections in queue.
 * Default is from empirical testing - in our case, more isn't necessarily better.
-* See https://gitlab.labs.nic.cz/knot/knot-resolver/-/merge_requests/968
+* See https://gitlab.nic.cz/knot/knot-resolver/-/merge_requests/968
 * */
 #ifndef TCP_BACKLOG_DEFAULT
 #define TCP_BACKLOG_DEFAULT 128
@@ -135,7 +137,7 @@ static int l_setuser(lua_State *L)
 /** Quit current executable. */
 static int l_quit(lua_State *L)
 {
-	engine_stop(engine_luaget(L));
+	engine_stop(the_worker->engine);
 	return 0;
 }
 
@@ -184,7 +186,7 @@ int engine_set_hostname(struct engine *engine, const char *hostname) {
 /** Return hostname. */
 static int l_hostname(lua_State *L)
 {
-	struct engine *engine = engine_luaget(L);
+	struct engine *engine = the_worker->engine;
 	if (lua_gettop(L) == 0) {
 		lua_pushstring(L, engine_get_hostname(engine));
 		return 1;
@@ -209,8 +211,7 @@ static int l_package_version(lua_State *L)
 /** Load root hints from zonefile. */
 static int l_hint_root_file(lua_State *L)
 {
-	struct engine *engine = engine_luaget(L);
-	struct kr_context *ctx = &engine->resolver;
+	struct kr_context *ctx = &the_worker->engine->resolver;
 	const char *file = lua_tostring(L, 1);
 
 	const char *err = engine_hint_root_file(ctx, file);
@@ -371,63 +372,6 @@ static int l_fromjson(lua_State *L)
 	return 1;
 }
 
-/** @internal Throw Lua error if expr is false */
-#define expr_checked(expr) \
-	if (!(expr)) { lua_pushboolean(L, false); lua_rawseti(L, -2, lua_objlen(L, -2) + 1); continue; }
-
-static int l_map(lua_State *L)
-{
-	/* We don't kr_log_deprecate() here for now.  Plan: after --forks gets *removed*,
-	 * kill internal uses of map() (e.g. from daf module) and add deprecation here.
-	 * Alternatively we might (attempt to) implement map() in another way. */
-	if (lua_gettop(L) != 1 || !lua_isstring(L, 1))
-		lua_error_p(L, "map('string with a lua expression')");
-
-	struct engine *engine = engine_luaget(L);
-	const char *cmd = lua_tostring(L, 1);
-	uint32_t len = strlen(cmd);
-	lua_newtable(L);
-
-	/* Execute on leader instance */
-	int ntop = lua_gettop(L);
-	engine_cmd(L, cmd, true);
-	lua_settop(L, ntop + 1); /* Push only one return value to table */
-	lua_rawseti(L, -2, 1);
-
-	for (size_t i = 0; i < engine->ipc_set.len; ++i) {
-		int fd = engine->ipc_set.at[i];
-		/* Send command */
-		expr_checked(write(fd, &len, sizeof(len)) == sizeof(len));
-		expr_checked(write(fd, cmd, len) == len);
-		/* Read response */
-		uint32_t rlen = 0;
-		if (read(fd, &rlen, sizeof(rlen)) == sizeof(rlen)) {
-			expr_checked(rlen < UINT32_MAX);
-			auto_free char *rbuf = malloc(rlen + 1);
-			expr_checked(rbuf != NULL);
-			expr_checked(read(fd, rbuf, rlen) == rlen);
-			rbuf[rlen] = '\0';
-			/* Unpack from JSON */
-			JsonNode *root_node = json_decode(rbuf);
-			if (root_node) {
-				l_unpack_json(L, root_node);
-			} else {
-				lua_pushlstring(L, rbuf, rlen);
-			}
-			json_delete(root_node);
-			lua_rawseti(L, -2, lua_objlen(L, -2) + 1);
-			continue;
-		}
-		/* Didn't respond */
-		lua_pushboolean(L, false);
-		lua_rawseti(L, -2, lua_objlen(L, -2) + 1);
-	}
-	return 1;
-}
-
-#undef expr_checked
-
-
 /*
  * Engine API.
  */
@@ -442,11 +386,13 @@ static int init_resolver(struct engine *engine)
 	engine->resolver.modules = &engine->modules;
 	engine->resolver.cache_rtt_tout_retry_interval = KR_NS_TIMEOUT_RETRY_INTERVAL;
 	/* Create OPT RR */
-	engine->resolver.opt_rr = mm_alloc(engine->pool, sizeof(knot_rrset_t));
-	if (!engine->resolver.opt_rr) {
+	engine->resolver.downstream_opt_rr = mm_alloc(engine->pool, sizeof(knot_rrset_t));
+	engine->resolver.upstream_opt_rr = mm_alloc(engine->pool, sizeof(knot_rrset_t));
+	if (!engine->resolver.downstream_opt_rr || !engine->resolver.upstream_opt_rr) {
 		return kr_error(ENOMEM);
 	}
-	knot_edns_init(engine->resolver.opt_rr, KR_EDNS_PAYLOAD, 0, KR_EDNS_VERSION, engine->pool);
+	knot_edns_init(engine->resolver.downstream_opt_rr, KR_EDNS_PAYLOAD, 0, KR_EDNS_VERSION, engine->pool);
+	knot_edns_init(engine->resolver.upstream_opt_rr, KR_EDNS_PAYLOAD, 0, KR_EDNS_VERSION, engine->pool);
 	/* Use default TLS padding */
 	engine->resolver.tls_padding = -1;
 	/* Empty init; filled via ./lua/postconfig.lua */
@@ -496,10 +442,6 @@ static int init_state(struct engine *engine)
 	lua_setglobal(engine->L, "tojson");
 	lua_pushcfunction(engine->L, l_fromjson);
 	lua_setglobal(engine->L, "fromjson");
-	lua_pushcfunction(engine->L, l_map);
-	lua_setglobal(engine->L, "map");
-	lua_pushlightuserdata(engine->L, engine);
-	lua_setglobal(engine->L, "__engine");
 	/* Random number generator */
 	lua_getfield(engine->L, LUA_GLOBALSINDEX, "math");
 	lua_getfield(engine->L, -1, "randomseed");
@@ -629,9 +571,6 @@ void engine_deinit(struct engine *engine)
 	 * e.g. the endpoint kind registry to work (inside ->net),
 	 * and this registry deinitization uses the lua state. */
 	network_close_force(&engine->net);
-	for (size_t i = 0; i < engine->ipc_set.len; ++i) {
-		close(engine->ipc_set.at[i]);
-	}
 	for (size_t i = 0; i < engine->modules.len; ++i) {
 		engine_unload(engine, engine->modules.at[i]);
 	}
@@ -650,7 +589,6 @@ void engine_deinit(struct engine *engine)
 	/* Free data structures */
 	array_clear(engine->modules);
 	array_clear(engine->backends);
-	array_clear(engine->ipc_set);
 	kr_ta_clear(&engine->resolver.trust_anchors);
 	kr_ta_clear(&engine->resolver.negative_anchors);
 	free(engine->hostname);
@@ -676,22 +614,6 @@ int engine_cmd(lua_State *L, const char *str, bool raw)
 	return engine_pcall(L, 2);
 }
 
-int engine_ipc(struct engine *engine, const char *expr)
-{
-	if (engine == NULL || engine->L == NULL) {
-		return kr_error(ENOEXEC);
-	}
-
-	/* Run expression and serialize response. */
-	engine_cmd(engine->L, expr, true);
-	if (lua_gettop(engine->L) > 0) {
-		l_tojson(engine->L);
-		return 1;
-	} else {
-		return 0;
-	}
-}
-
 int engine_load_sandbox(struct engine *engine)
 {
 	/* Init environment */
@@ -858,11 +780,3 @@ int engine_unregister(struct engine *engine, const char *name)
 	return kr_error(ENOENT);
 }
 
-struct engine *engine_luaget(lua_State *L)
-{
-	lua_getglobal(L, "__engine");
-	struct engine *engine = lua_touserdata(L, -1);
-	if (!engine) luaL_error(L, "internal error, empty engine pointer");
-	lua_pop(L, 1);
-	return engine;
-}
