@@ -1,18 +1,28 @@
-/*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2014-2020 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
  *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
-#include <string.h>
-#include <libknot/errcode.h>
+#include "daemon/io.h"
+
+#include <assert.h>
 #include <contrib/ucw/lib.h>
 #include <contrib/ucw/mempool.h>
-#include <assert.h>
+#include <libknot/errcode.h>
+#include <string.h>
+#include <sys/resource.h>
+
+#if ENABLE_XDP
+	#include <libknot/xdp/xdp.h>
+	#include <net/if.h>
+#endif
 
-#include "daemon/io.h"
 #include "daemon/network.h"
 #include "daemon/worker.h"
 #include "daemon/tls.h"
+#include "daemon/http.h"
 #include "daemon/session.h"
+#include "contrib/cleanup.h"
+#include "lib/utils.h"
 
 #define negotiate_bufsize(func, handle, bufsize_want) do { \
     int bufsize = 0; (func)((handle), &bufsize); \
@@ -60,21 +70,10 @@ static void handle_getbuf(uv_handle_t* handle, size_t suggested_size, uv_buf_t*
 void udp_recv(uv_udp_t *handle, ssize_t nread, const uv_buf_t *buf,
 	const struct sockaddr *addr, unsigned flags)
 {
-	uv_loop_t *loop = handle->loop;
-	struct worker_ctx *worker = loop->data;
 	struct session *s = handle->data;
-	if (session_flags(s)->closing) {
+	if (session_flags(s)->closing || nread <= 0 || addr->sa_family == AF_UNSPEC)
 		return;
-	}
-	if (nread <= 0) {
-		if (nread < 0) { /* Error response, notify resolver */
-			worker_submit(s, NULL, NULL);
-		} /* nread == 0 is for freeing buffers, we don't need to do this */
-		return;
-	}
-	if (addr->sa_family == AF_UNSPEC) {
-		return;
-	}
+
 	if (session_flags(s)->outgoing) {
 		const struct sockaddr *peer = session_get_peer(s);
 		assert(peer->sa_family != AF_UNSPEC);
@@ -89,7 +88,7 @@ void udp_recv(uv_udp_t *handle, ssize_t nread, const uv_buf_t *buf,
 	assert(consumed == nread); (void)consumed;
 	session_wirebuf_process(s, addr);
 	session_wirebuf_discard(s);
-	mp_flush(worker->pkt_pool.ctx);
+	mp_flush(the_worker->pkt_pool.ctx);
 }
 
 static int family_to_freebind_option(sa_family_t sa_family, int *level, int *name)
@@ -153,6 +152,23 @@ int io_bind(const struct sockaddr *addr, int type, const endpoint_flags_t *flags
 			if (setsockopt(fd, optlevel, optname, &yes, sizeof(yes)))
 				return kr_error(errno);
 		}
+
+		/* Linux 3.15 has IP_PMTUDISC_OMIT which makes sockets
+		 * ignore PMTU information and send packets with DF=0.
+		 * This mitigates DNS fragmentation attacks by preventing
+		 * forged PMTU information.  FreeBSD already has same semantics
+		 * without setting the option.
+			https://gitlab.nic.cz/knot/knot-dns/-/issues/640
+		 */
+#if defined(IP_MTU_DISCOVER) && defined(IP_PMTUDISC_OMIT)
+		int omit = IP_PMTUDISC_OMIT;
+		if (type == SOCK_DGRAM && addr->sa_family == AF_INET
+		    && setsockopt(fd, IPPROTO_IP, IP_MTU_DISCOVER, &omit, sizeof(omit))) {
+			kr_log_error(
+				"[ io ] failed to disable Path MTU discovery for %s UDP: %s\n",
+				kr_straddr(addr), strerror(errno));
+		}
+#endif
 	}
 
 	if (bind(fd, addr, kr_sockaddr_len(addr)))
@@ -175,7 +191,7 @@ int io_listen_udp(uv_loop_t *loop, uv_udp_t *handle, int fd)
 	uv_handle_t *h = (uv_handle_t *)handle;
 	check_bufsize(h);
 	/* Handle is already created, just create context. */
-	struct session *s = session_new(h, false);
+	struct session *s = session_new(h, false, false);
 	assert(s);
 	session_flags(s)->outgoing = false;
 
@@ -195,11 +211,9 @@ void tcp_timeout_trigger(uv_timer_t *timer)
 
 	assert(!session_flags(s)->closing);
 
-	struct worker_ctx *worker = timer->loop->data;
-
 	if (!session_tasklist_is_empty(s)) {
 		int finalized = session_tasklist_finalize_expired(s);
-		worker->stats.timeout += finalized;
+		the_worker->stats.timeout += finalized;
 		/* session_tasklist_finalize_expired() may call worker_task_finalize().
 		 * If session is a source session and there were IO errors,
 		 * worker_task_finalize() can filnalize all tasks and close session. */
@@ -220,13 +234,12 @@ void tcp_timeout_trigger(uv_timer_t *timer)
 			struct qr_task *t = session_waitinglist_pop(s, false);
 			worker_task_finalize(t, KR_STATE_FAIL);
 			worker_task_unref(t);
-			worker->stats.timeout += 1;
+			the_worker->stats.timeout += 1;
 			if (session_flags(s)->closing) {
 				return;
 			}
 		}
-		const struct engine *engine = worker->engine;
-		const struct network *net = &engine->net;
+		const struct network *net = &the_worker->engine->net;
 		uint64_t idle_in_timeout = net->tcp.in_idle_timeout;
 		uint64_t last_activity = session_last_activity(s);
 		uint64_t idle_time = kr_now() - last_activity;
@@ -241,8 +254,8 @@ void tcp_timeout_trigger(uv_timer_t *timer)
 			kr_log_verbose("[io] => closing connection to '%s'\n",
 				       peer_str ? peer_str : "");
 			if (session_flags(s)->outgoing) {
-				worker_del_tcp_waiting(worker, peer);
-				worker_del_tcp_connected(worker, peer);
+				worker_del_tcp_waiting(the_worker, peer);
+				worker_del_tcp_connected(the_worker, peer);
 			}
 			session_close(s);
 		}
@@ -300,6 +313,26 @@ static void tcp_recv(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf)
 		data = session_wirebuf_get_free_start(s);
 		data_len = consumed;
 	}
+#if ENABLE_DOH2
+	if (session_flags(s)->has_http) {
+		consumed = http_process_input_data(s, data, data_len);
+		if (consumed < 0) {
+			if (kr_verbose_status) {
+				struct sockaddr *peer = session_get_peer(s);
+				char *peer_str = kr_straddr(peer);
+				kr_log_verbose("[io] => connection to '%s': "
+				       "error processing HTTP data, close\n",
+				       peer_str ? peer_str : "");
+			}
+			worker_end_tcp(s);
+			return;
+		} else if (consumed == 0) {
+			return;
+		}
+		data = session_wirebuf_get_free_start(s);
+		data_len = consumed;
+	}
+#endif
 
 	/* data points to start of the free space in session wire buffer.
 	   Simple increase internal counter. */
@@ -312,11 +345,27 @@ static void tcp_recv(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf)
 		worker_end_tcp(s);
 	}
 	session_wirebuf_compress(s);
-	struct worker_ctx *worker = handle->loop->data;
-	mp_flush(worker->pkt_pool.ctx);
+	mp_flush(the_worker->pkt_pool.ctx);
 }
 
-static void _tcp_accept(uv_stream_t *master, int status, bool tls)
+#if ENABLE_DOH2
+static ssize_t tls_send(const uint8_t *buf, const size_t len, struct session *session)
+{
+	struct tls_ctx *ctx = session_tls_get_server_ctx(session);
+	ssize_t sent = 0;
+	assert(ctx);
+
+	sent = gnutls_record_send(ctx->c.tls_session, buf, len);
+	if (sent < 0) {
+		kr_log_verbose("[http] gnutls_record_send failed: %s (%zd)\n",
+			       gnutls_strerror_name(sent), sent);
+		return kr_error(EIO);
+	}
+	return sent;
+}
+#endif
+
+static void _tcp_accept(uv_stream_t *master, int status, bool tls, bool http)
 {
  	if (status != 0) {
 		return;
@@ -328,7 +377,7 @@ static void _tcp_accept(uv_stream_t *master, int status, bool tls)
 		return;
 	}
 	int res = io_create(master->loop, (uv_handle_t *)client,
-			    SOCK_STREAM, AF_UNSPEC, tls);
+			    SOCK_STREAM, AF_UNSPEC, tls, http);
 	if (res) {
 		if (res == UV_EMFILE) {
 			worker->too_many_open = true;
@@ -380,7 +429,7 @@ static void _tcp_accept(uv_stream_t *master, int status, bool tls)
 	uint64_t timeout = KR_CONN_RTT_MAX / 2;
 	if (tls) {
 		timeout += TLS_MAX_HANDSHAKE_TIME;
-		struct tls_ctx_t *ctx = session_tls_get_server_ctx(s);
+		struct tls_ctx *ctx = session_tls_get_server_ctx(s);
 		if (!ctx) {
 			ctx = tls_new(worker);
 			if (!ctx) {
@@ -389,32 +438,94 @@ static void _tcp_accept(uv_stream_t *master, int status, bool tls)
 			}
 			ctx->c.session = s;
 			ctx->c.handshake_state = TLS_HS_IN_PROGRESS;
+
+			/* Configure ALPN. */
+			gnutls_datum_t proto;
+			if (!http) {
+				proto.data = (unsigned char *)"dot";
+				proto.size = 3;
+			} else {
+				proto.data = (unsigned char *)"h2";
+				proto.size = 2;
+			}
+			unsigned int flags = 0;
+#if GNUTLS_VERSION_NUMBER >= 0x030500
+			/* Mandatory ALPN means the protocol must match if and
+			 * only if ALPN extension is used by the client. */
+			flags |= GNUTLS_ALPN_MANDATORY;
+#endif
+			ret = gnutls_alpn_set_protocols(ctx->c.tls_session, &proto, 1, flags);
+			if (ret != GNUTLS_E_SUCCESS) {
+				session_close(s);
+				return;
+			}
+
 			session_tls_set_server_ctx(s, ctx);
 		}
 	}
+#if ENABLE_DOH2
+	if (http) {
+		struct http_ctx *ctx = session_http_get_server_ctx(s);
+		if (!ctx) {
+			if (!tls) {  /* Plain HTTP is not supported. */
+				session_close(s);
+				return;
+			}
+			ctx = http_new(s, tls_send);
+			if (!ctx) {
+				session_close(s);
+				return;
+			}
+			session_http_set_server_ctx(s, ctx);
+		}
+	}
+#endif
 	session_timer_start(s, tcp_timeout_trigger, timeout, idle_in_timeout);
 	io_start_read((uv_handle_t *)client);
 }
 
 static void tcp_accept(uv_stream_t *master, int status)
 {
-	_tcp_accept(master, status, false);
+	_tcp_accept(master, status, false, false);
 }
 
 static void tls_accept(uv_stream_t *master, int status)
 {
-	_tcp_accept(master, status, true);
+	_tcp_accept(master, status, true, false);
 }
 
-int io_listen_tcp(uv_loop_t *loop, uv_tcp_t *handle, int fd, int tcp_backlog, bool has_tls)
+#if ENABLE_DOH2
+static void https_accept(uv_stream_t *master, int status)
 {
-	const uv_connection_cb connection = has_tls ? tls_accept : tcp_accept;
+	_tcp_accept(master, status, true, true);
+}
+#endif
+
+int io_listen_tcp(uv_loop_t *loop, uv_tcp_t *handle, int fd, int tcp_backlog, bool has_tls, bool has_http)
+{
+	uv_connection_cb connection;
+
 	if (!handle) {
 		return kr_error(EINVAL);
 	}
 	int ret = uv_tcp_init(loop, handle);
 	if (ret) return ret;
 
+	if (has_tls && has_http) {
+#if ENABLE_DOH2
+		connection = https_accept;
+#else
+		kr_log_error("[ io ] kresd was compiled without libnghttp2 support\n");
+		return kr_error(ENOPROTOOPT);
+#endif
+	} else if (has_tls) {
+		connection = tls_accept;
+	} else if (has_http) {
+		return kr_error(EPROTONOSUPPORT);
+	} else {
+		connection = tcp_accept;
+	}
+
 	ret = uv_tcp_open(handle, (uv_os_sock_t) fd);
 	if (ret) return ret;
 
@@ -450,6 +561,19 @@ int io_listen_tcp(uv_loop_t *loop, uv_tcp_t *handle, int fd, int tcp_backlog, bo
 	return 0;
 }
 
+
+enum io_stream_mode {
+	io_mode_text = 0,
+	io_mode_binary = 1,
+};
+
+struct io_stream_data {
+	enum io_stream_mode mode;
+	size_t blen; ///< length of `buf`
+	char *buf;  ///< growing buffer residing on `pool` (mp_append_*)
+	knot_mm_t *pool;
+};
+
 /**
  * TTY control: process input and free() the buffer.
  *
@@ -459,114 +583,145 @@ int io_listen_tcp(uv_loop_t *loop, uv_tcp_t *handle, int fd, int tcp_backlog, bo
  */
 void io_tty_process_input(uv_stream_t *stream, ssize_t nread, const uv_buf_t *buf)
 {
-	char *commands = buf ? buf->base : NULL; /* To be free()d on return. */
+	auto_free char *commands = buf ? buf->base : NULL;
 
 	/* Set output streams */
 	FILE *out = stdout;
-	uv_os_fd_t stream_fd = 0;
+	uv_os_fd_t stream_fd = -1;
 	struct args *args = the_args;
-	if (uv_fileno((uv_handle_t *)stream, &stream_fd)) {
+	struct io_stream_data *data = (struct io_stream_data*) stream->data;
+	if (nread < 0 || uv_fileno((uv_handle_t *)stream, &stream_fd)) {
+		mp_delete(data->pool->ctx);
 		uv_close((uv_handle_t *)stream, (uv_close_cb) free);
-		free(commands);
+		return;
+	}
+	if (nread <= 0) {
 		return;
 	}
 	if (stream_fd != STDIN_FILENO) {
-		if (nread < 0) { /* Close if disconnected */
-			uv_close((uv_handle_t *)stream, (uv_close_cb) free);
-		}
-		if (nread <= 0) {
-			free(commands);
-			return;
-		}
 		uv_os_fd_t dup_fd = dup(stream_fd);
 		if (dup_fd >= 0) {
 			out = fdopen(dup_fd, "w");
 		}
 	}
 
-	char *cmd = NULL;
-	/* Execute */
-	if (stream && commands && nread > 0) {
-		/* Ensure commands is 0-terminated */
-		if (commands[nread - 1] == '\n') {
-			commands[nread - 1] = '\0';
-		} else {
-			if (nread >= buf->len) { /* only equality should be possible */
-				char *newbuf = realloc(commands, nread + 1);
-				if (!newbuf)
-					goto finish;
-				commands = newbuf;
-			}
-			commands[nread] = '\0';
-		}
+	/** The current single command and the remaining command(s). */
+	char *cmd, *cmd_next = NULL;
+	bool incomplete_cmd = false;
 
-		const char *delim = args->quiet ? "" : "> ";
-
-		/* No command, just new line */
-		if (nread == 1 && args->tty_binary_output == false && commands[nread-1] == '\0') {
-			if (stream_fd != STDIN_FILENO) {
-				fprintf(out, "%s", delim);
-			}
-			if (stream_fd == STDIN_FILENO || VERBOSE_STATUS) {
-				fprintf(stdout, "%s", delim);
-			}
-		}
-
-		cmd = strtok(commands, "\n");
-		while (cmd != NULL) {
-			/* Pseudo-command for switching to "binary output"; */
-			if (strcmp(cmd, "__binary") == 0) {
-				args->tty_binary_output = true;
-				cmd = strtok(NULL, "\n");
-				continue;
-			}
-
-			lua_State *L = the_worker->engine->L;
-			int ret = engine_cmd(L, cmd, false);
-			const char *message = "";
-			if (lua_gettop(L) > 0) {
-				message = lua_tostring(L, -1);
-			}
-
-			/* Simpler output in binary mode */
-			if (args->tty_binary_output) {
-				size_t len_s = strlen(message);
-				if (len_s > UINT32_MAX) {
-					cmd = strtok(NULL, "\n");
-					continue;
-				}
-				uint32_t len_n = htonl(len_s);
-				fwrite(&len_n, sizeof(len_n), 1, out);
-				fwrite(message, len_s, 1, out);
-				lua_settop(L, 0);
-				cmd = strtok(NULL, "\n");
-				continue;
-			}
-			/* Log to remote socket if connected */
-			if (stream_fd != STDIN_FILENO) {
-				if (VERBOSE_STATUS)
-					fprintf(stdout, "%s\n", cmd); /* Duplicate command to logs */
-				if (message)
-					fprintf(out, "%s", message); /* Duplicate output to sender */
-				if (message || !args->quiet)
-					fprintf(out, "\n");
-				fprintf(out, "%s", delim);
-			}
-			if (stream_fd == STDIN_FILENO || VERBOSE_STATUS) {
-				/* Log to standard streams */
-				FILE *fp_out = ret ? stderr : stdout;
-				if (message)
-					fprintf(fp_out, "%s", message);
-				if (message || !args->quiet)
-					fprintf(fp_out, "\n");
-				fprintf(fp_out, "%s", delim);
-			}
-			lua_settop(L, 0);
-			cmd = strtok(NULL, "\n");
-		}
+	if (!(stream && commands && nread > 0)) {
+		goto finish;
 	}
+	/* Execute */
+
+	if (commands[nread - 1] != '\n') {
+		incomplete_cmd = true;
+	}
+	/* Ensure commands is 0-terminated */
+	if (nread >= buf->len) { /* only equality should be possible */
+		char *newbuf = realloc(commands, nread + 1);
+		if (!newbuf)
+			goto finish;
+		commands = newbuf;
+	}
+	commands[nread] = '\0';
+
+	char *boundary = "\n\0";
+	cmd = strtok(commands, "\n");
+	/* strtok skip '\n' but we need process alone '\n' too */
+	if (commands[0] == '\n') {
+		cmd_next = cmd;
+		cmd = boundary;
+	} else {
+		cmd_next = strtok(NULL, "\n");
+	}
+
+	/** Moving pointer to end of buffer with incomplete command. */
+	char *pbuf = data->buf + data->blen;
+	lua_State *L = the_worker->engine->L;
+	while (cmd != NULL) {
+		/* Last command is incomplete - save it and execute later */
+		if (incomplete_cmd && cmd_next == NULL) {
+			pbuf = mp_append_string(data->pool->ctx, pbuf, cmd);
+			mp_append_char(data->pool->ctx, pbuf, '\0');
+			data->buf = mp_ptr(data->pool->ctx);
+			data->blen = data->blen + strlen(cmd);
+
+			/* There is new incomplete command */
+			if (commands[nread - 1] == '\n')
+				incomplete_cmd = false;
+			goto next_iter;
+		}
+
+		/* Process incomplete command from previously call */
+		if (data->blen > 0) {
+			if (commands[0] != '\n' && commands[0] != '\0') {
+				pbuf = mp_append_string(data->pool->ctx, pbuf, cmd);
+				mp_append_char(data->pool->ctx, pbuf, '\0');
+				data->buf = mp_ptr(data->pool->ctx);
+				cmd = data->buf;
+			} else {
+				cmd = data->buf;
+			}
+			data->blen = 0;
+			pbuf = data->buf;
+		}
+
+		/* Pseudo-command for switching to "binary output"; */
+		if (strcmp(cmd, "__binary") == 0) {
+			data->mode = io_mode_binary;
+			goto next_iter;
+		}
+
+		int ret = engine_cmd(L, cmd, false);
+		const char *message = NULL;
+		size_t len_s;
+		if (lua_gettop(L) > 0) {
+			message = lua_tolstring(L, -1, &len_s);
+		}
+
+		/* Simpler output in binary mode */
+		if (data->mode == io_mode_binary) {
+			/* Leader expects length field in all cases */
+			if (!message || len_s > UINT32_MAX) {
+				kr_log_error("[io] unrepresentable respose on control socket, "
+						"sending back empty block (command '%s')\n", cmd);
+				len_s = 0;
+			}
+			uint32_t len_n = htonl(len_s);
+			fwrite(&len_n, sizeof(len_n), 1, out);
+			if (len_s > 0)
+				fwrite(message, len_s, 1, out);
+			goto next_iter;
+		}
+
+		/* Log to remote socket if connected */
+		const char *delim = args->quiet ? "" : "> ";
+		if (stream_fd != STDIN_FILENO) {
+			if (VERBOSE_STATUS)
+				fprintf(stdout, "%s\n", cmd); /* Duplicate command to logs */
+			if (message)
+				fprintf(out, "%s", message); /* Duplicate output to sender */
+			if (message || !args->quiet)
+				fprintf(out, "\n");
+			fprintf(out, "%s", delim);
+		}
+		if (stream_fd == STDIN_FILENO || VERBOSE_STATUS) {
+			/* Log to standard streams */
+			FILE *fp_out = ret ? stderr : stdout;
+			if (message)
+				fprintf(fp_out, "%s", message);
+			if (message && !args->quiet)
+				fprintf(fp_out, "\n");
+			fprintf(fp_out, "%s", delim);
+		}
+	next_iter:
+		lua_settop(L, 0); /* not required in some cases but harmless */
+		cmd = cmd_next;
+		cmd_next = strtok(NULL, "\n");
+	}
+
 finish:
-	free(commands);
 	/* Close if redirected */
 	if (stream_fd != STDIN_FILENO) {
 		fclose(out);
@@ -579,14 +734,39 @@ void io_tty_alloc(uv_handle_t *handle, size_t suggested, uv_buf_t *buf)
 	buf->base = malloc(suggested);
 }
 
+struct io_stream_data *io_tty_alloc_data() {
+	knot_mm_t _pool = {
+		.ctx = mp_new(4096),
+		.alloc = (knot_mm_alloc_t) mp_alloc,
+	};
+	knot_mm_t *pool = mm_alloc(&_pool, sizeof(*pool));
+	if (!pool) {
+		return NULL;
+	}
+	memcpy(pool, &_pool, sizeof(*pool));
+
+	struct io_stream_data *data = mm_alloc(pool, sizeof(struct io_stream_data));
+
+	data->buf = mp_start(pool->ctx, 512);
+	data->mode = io_mode_text;
+	data->blen = 0;
+	data->pool = pool;
+
+	return data;
+}
+
 void io_tty_accept(uv_stream_t *master, int status)
 {
-	uv_tcp_t *client = malloc(sizeof(*client));
+	struct io_stream_data *data = io_tty_alloc_data();
+	/* We can't use any allocations after mp_start() and it's easier anyway. */
+	uv_pipe_t *client = malloc(sizeof(*client));
+	client->data = data;
+
 	struct args *args = the_args;
-	if (client) {
-		 uv_tcp_init(master->loop, client);
+	if (client && client->data) {
+		 uv_pipe_init(master->loop, client, 0);
 		 if (uv_accept(master, (uv_stream_t *)client) != 0) {
-			free(client);
+			mp_delete(data->pool->ctx);
 			return;
 		 }
 		 uv_read_start((uv_stream_t *)client, io_tty_alloc, io_tty_process_input);
@@ -617,7 +797,130 @@ int io_listen_pipe(uv_loop_t *loop, uv_pipe_t *handle, int fd)
 	return 0;
 }
 
-int io_create(uv_loop_t *loop, uv_handle_t *handle, int type, unsigned family, bool has_tls)
+#if ENABLE_XDP
+static void xdp_rx(uv_poll_t* handle, int status, int events)
+{
+	const int XDP_RX_BATCH_SIZE = 64;
+	if (status < 0) {
+		kr_log_error("[xdp] poll status %d: %s\n", status, uv_strerror(status));
+		return;
+	}
+	if (events != UV_READABLE) {
+		kr_log_error("[xdp] poll unexpected events: %d\n", events);
+		return;
+	}
+
+	xdp_handle_data_t *xhd = handle->data;
+	assert(xhd && xhd->session && xhd->socket);
+	uint32_t rcvd;
+	knot_xdp_msg_t msgs[XDP_RX_BATCH_SIZE];
+	int ret = knot_xdp_recv(xhd->socket, msgs, XDP_RX_BATCH_SIZE, &rcvd);
+	if (ret == KNOT_EOK) {
+		kr_log_verbose("[xdp] poll triggered, processing a batch of %d packets\n",
+			(int)rcvd);
+	} else {
+		kr_log_error("[xdp] knot_xdp_recv(): %d, %s\n", ret, knot_strerror(ret));
+		assert(false); // ATM it can only be returned when called incorrectly
+		return;
+	}
+	assert(rcvd <= XDP_RX_BATCH_SIZE);
+	for (int i = 0; i < rcvd; ++i) {
+		const knot_xdp_msg_t *msg = &msgs[i];
+		assert(msg->payload.iov_len <= KNOT_WIRE_MAX_PKTSIZE);
+		knot_pkt_t *kpkt = knot_pkt_new(msg->payload.iov_base, msg->payload.iov_len,
+						&the_worker->pkt_pool);
+		if (kpkt == NULL) {
+			ret = kr_error(ENOMEM);
+		} else {
+			ret = worker_submit(xhd->session,
+					(const struct sockaddr *)&msg->ip_from,
+					(const struct sockaddr *)&msg->ip_to,
+					msg->eth_from, msg->eth_to, kpkt);
+		}
+		if (ret)
+			kr_log_verbose("[xdp] worker_submit() == %d: %s\n", ret, kr_strerror(ret));
+		mp_flush(the_worker->pkt_pool.ctx);
+	}
+	knot_xdp_recv_finish(xhd->socket, msgs, rcvd);
+}
+/// Warn if the XDP program is running in emulated mode (XDP_SKB)
+static void xdp_warn_mode(const char *ifname)
+{
+	assert(ifname);
+	const unsigned if_index = if_nametoindex(ifname);
+	if (!if_index) {
+		kr_log_info("[xdp] warning: interface %s, unexpected error when converting its name: %s\n",
+				ifname, strerror(errno));
+		return;
+	}
+
+	const knot_xdp_mode_t mode = knot_eth_xdp_mode(if_index);
+	switch (mode) {
+	case KNOT_XDP_MODE_FULL:
+		return;
+	case KNOT_XDP_MODE_EMUL:
+		kr_log_info("[xdp] warning: interface %s running only with XDP emulation\n",
+				ifname);
+		return;
+	case KNOT_XDP_MODE_NONE: // enum warnings from compiler
+		break;
+	}
+	kr_log_info("[xdp] warning: interface %s running in unexpected XDP mode %d\n",
+			ifname, (int)mode);
+}
+int io_listen_xdp(uv_loop_t *loop, struct endpoint *ep, const char *ifname)
+{
+	if (!ep || !ep->handle) {
+		return kr_error(EINVAL);
+	}
+
+	// RLIMIT_MEMLOCK often needs raising when operating on BPF
+	static int ret_limit = 1;
+	if (ret_limit == 1) {
+		struct rlimit no_limit = { RLIM_INFINITY, RLIM_INFINITY };
+		ret_limit = setrlimit(RLIMIT_MEMLOCK, &no_limit)
+			? kr_error(errno) : 0;
+	}
+	if (ret_limit) return ret_limit;
+
+	xdp_handle_data_t *xhd = malloc(sizeof(*xhd));
+	if (!xhd) return kr_error(ENOMEM);
+
+	const int port = ep->port ? ep->port : KNOT_XDP_LISTEN_PORT_ALL;
+	xhd->socket = NULL; // needed for some reason
+	int ret = knot_xdp_init(&xhd->socket, ifname, ep->nic_queue, port,
+				KNOT_XDP_LOAD_BPF_MAYBE);
+	if (!ret) xdp_warn_mode(ifname);
+
+	if (!ret) ret = uv_idle_init(loop, &xhd->tx_waker);
+	if (ret) {
+		free(xhd);
+		return kr_error(ret);
+	}
+	assert(xhd->socket);
+	xhd->tx_waker.data = xhd->socket;
+
+	ep->fd = knot_xdp_socket_fd(xhd->socket); // probably not useful
+	ret = uv_poll_init(loop, (uv_poll_t *)ep->handle, ep->fd);
+	if (ret) {
+		knot_xdp_deinit(xhd->socket);
+		free(xhd);
+		return kr_error(ret);
+	}
+
+	// beware: this sets poll_handle->data
+	xhd->session = session_new(ep->handle, false, false);
+	assert(!session_flags(xhd->session)->outgoing);
+	session_get_sockname(xhd->session)->sa_family = AF_XDP; // to have something in there
+
+	ep->handle->data = xhd;
+	ret = uv_poll_start((uv_poll_t *)ep->handle, UV_READABLE, xdp_rx);
+	return ret;
+}
+#endif
+
+
+int io_create(uv_loop_t *loop, uv_handle_t *handle, int type, unsigned family, bool has_tls, bool has_http)
 {
 	int ret = -1;
 	if (type == SOCK_DGRAM) {
@@ -629,21 +932,31 @@ int io_create(uv_loop_t *loop, uv_handle_t *handle, int type, unsigned family, b
 	if (ret != 0) {
 		return ret;
 	}
-	struct session *s = session_new(handle, has_tls);
+	struct session *s = session_new(handle, has_tls, has_http);
 	if (s == NULL) {
 		ret = -1;
 	}
 	return ret;
 }
 
-void io_deinit(uv_handle_t *handle)
+static void io_deinit(uv_handle_t *handle)
 {
-	if (!handle) {
+	if (!handle || !handle->data) {
 		return;
 	}
-	if (handle->data) {
+	if (handle->type != UV_POLL) {
 		session_free(handle->data);
-		handle->data = NULL;
+	} else {
+	#if ENABLE_XDP
+		xdp_handle_data_t *xhd = handle->data;
+		uv_idle_stop(&xhd->tx_waker);
+		uv_close((uv_handle_t *)&xhd->tx_waker, NULL);
+		session_free(xhd->session);
+		knot_xdp_deinit(xhd->socket);
+		free(xhd);
+	#else
+		assert(false);
+	#endif
 	}
 }
 
