@@ -455,6 +455,27 @@ ffi.metatype( knot_rrset_t, {
 			C.free(dump[0])
 			return result
 		end,
+		txt_fields = function(rr, i)
+			assert(ffi.istype(knot_rrset_t, rr))
+			assert(i >= 0 and i < rr:rdcount())
+			local bufsize = 1024
+			local dump = ffi.new('char *', C.malloc(bufsize))
+			ffi.gc(dump, C.free)
+
+			local ret = knot.knot_rrset_txt_dump_data(rr, i, dump, 1024,
+							knot.KNOT_DUMP_STYLE_DEFAULT)
+			if ret >= 0 then
+				local out = {}
+				out.owner = dname2str(rr:owner())
+				out.ttl = rr:ttl()
+				out.class = kres.tostring.class[rr:class()]
+				out.type = kres.tostring.type[rr.type]
+				out.rdata = ffi.string(dump, ret)
+				return out
+			else
+				panic('knot_rrset_txt_dump_data failure ' .. tostring(ret))
+			end
+		end,
 		-- Return RDATA count for this RR set
 		rdcount = function(rr)
 			assert(ffi.istype(knot_rrset_t, rr))
@@ -893,6 +914,11 @@ ffi.metatype( kr_request_t, {
 			req.vars_ref = ref
 			return var
 		end,
+		-- Ensure that answer exists and return it; can't fail.
+		ensure_answer = function (req)
+			assert(ffi.istype(kr_request_t, req))
+			return C.kr_request_ensure_answer(req)
+		end,
 	},
 })
 
@@ -904,11 +930,21 @@ local function c_array_iter(t, i)
 end
 
 -- Metatype for a single ranked record array entry (one RRset)
+local function rank_tostring(rank)
+	local names = {}
+	for name, value in pairs(const_rank) do
+		if ffi.C.kr_rank_test(rank, value) then
+			table.insert(names, string.lower(name))
+		end
+	end
+	return string.format('0%.2o (%s)', rank, table.concat(names, ' '))
+end
+
 local ranked_rr_array_entry_t = ffi.typeof('ranked_rr_array_entry_t')
 ffi.metatype(ranked_rr_array_entry_t, {
 	__tostring = function(self)
-		return string.format('; ranked rrset to_wire %s, rank 0%.2o, cached %s, qry_uid %s, revalidations %s\n%s',
-		self.to_wire, self.rank, self.cached, self.qry_uid,
+		return string.format('; ranked rrset to_wire %s, rank %s, cached %s, qry_uid %s, revalidations %s\n%s',
+		self.to_wire, rank_tostring(self.rank), self.cached, self.qry_uid,
 		self.revalidation_cnt, string.format('%s', self.rr))
 	end
 })
@@ -1061,7 +1097,7 @@ kres = {
 		if ret ~= 1 then return nil end
 		return ffi.string(addr_buf, C.kr_family_len(family))
 	end,
-	context = function () return ffi.cast('struct kr_context *', __engine) end,
+	context = function () return ffi.C.the_worker.engine.resolver end,
 
 	knot_pkt_rr = knot_pkt_rr,
 }
