@@ -28,13 +28,13 @@
 #include <sys/resource.h>
 #include <unistd.h>
 
-#ifdef ENABLE_CAP_NG
+#if ENABLE_CAP_NG
 #include <cap-ng.h>
 #endif
 
 #include <lua.h>
 #include <uv.h>
-#if SYSTEMD_VERSION > 0
+#if ENABLE_LIBSYSTEMD
 #include <systemd/sd-daemon.h>
 #endif
 #include <libknot/error.h>
@@ -42,97 +42,6 @@
 
 struct args the_args_value;  /** Static allocation for the_args singleton. */
 
-
-/* @internal AF_LOCAL reads may still be interrupted, loop it. */
-static bool ipc_readall(int fd, char *dst, size_t len)
-{
-	while (len > 0) {
-		int rb = read(fd, dst, len);
-		if (rb > 0) {
-			dst += rb;
-			len -= rb;
-		} else if (errno != EAGAIN && errno != EINTR) {
-			return false;
-		}
-	}
-	return true;
-}
-
-static void ipc_activity(uv_poll_t *handle, int status, int events)
-{
-	struct engine *engine = handle->data;
-	if (status != 0) {
-		kr_log_error("[system] ipc: %s\n", uv_strerror(status));
-		return;
-	}
-	/* Get file descriptor from handle */
-	uv_os_fd_t fd = 0;
-	(void) uv_fileno((uv_handle_t *)(handle), &fd);
-	/* Read expression from IPC pipe */
-	uint32_t len = 0;
-	auto_free char *rbuf = NULL;
-	if (!ipc_readall(fd, (char *)&len, sizeof(len))) {
-		goto failure;
-	}
-	if (len < UINT32_MAX) {
-		rbuf = malloc(len + 1);
-	} else {
-		errno = EINVAL;
-	}
-	if (!rbuf) {
-		goto failure;
-	}
-	if (!ipc_readall(fd, rbuf, len)) {
-		goto failure;
-	}
-	rbuf[len] = '\0';
-	/* Run expression */
-	const char *message = "";
-	int ret = engine_ipc(engine, rbuf);
-	if (ret > 0) {
-		message = lua_tostring(engine->L, -1);
-	}
-	/* Clear the Lua stack */
-	lua_settop(engine->L, 0);
-	/* Send response back */
-	len = strlen(message);
-	if (write(fd, &len, sizeof(len)) != sizeof(len) ||
-		write(fd, message, len) != len) {
-		goto failure;
-	}
-	return; /* success! */
-failure:
-	/* Note that if the piped command got read or written partially,
-	 * we would get out of sync and only receive rubbish now.
-	 * Therefore we prefer to stop IPC, but we try to continue with all else.
-	 */
-	kr_log_error("[system] stopping ipc because of: %s\n", strerror(errno));
-	uv_poll_stop(handle);
-	uv_close((uv_handle_t *)handle, (uv_close_cb)free);
-}
-
-static bool ipc_watch(uv_loop_t *loop, struct engine *engine, int fd)
-{
-	uv_poll_t *poller = malloc(sizeof(*poller));
-	if (!poller) {
-		return false;
-	}
-	int ret = uv_poll_init(loop, poller, fd);
-	if (ret != 0) {
-		free(poller);
-		return false;
-	}
-	poller->data = engine;
-	ret = uv_poll_start(poller, UV_READABLE, ipc_activity);
-	if (ret != 0) {
-		free(poller);
-		return false;
-	}
-	/* libuv sets O_NONBLOCK whether we want it or not */
-	(void) fcntl(fd, F_SETFD, fcntl(fd, F_GETFL) & ~O_NONBLOCK);
-	return true;
-}
-
 static void signal_handler(uv_signal_t *handle, int signum)
 {
 	uv_stop(uv_default_loop());
@@ -180,15 +89,10 @@ end:
  * Server operation.
  */
 
-static int fork_workers(fd_array_t *ipc_set, int forks)
+static int fork_workers(int forks)
 {
 	/* Fork subprocesses if requested */
 	while (--forks > 0) {
-		int sv[2] = {-1, -1};
-		if (socketpair(AF_LOCAL, SOCK_STREAM, 0, sv) < 0) {
-			perror("[system] socketpair");
-			return kr_error(errno);
-		}
 		int pid = fork();
 		if (pid < 0) {
 			perror("[system] fork");
@@ -197,16 +101,7 @@ static int fork_workers(fd_array_t *ipc_set, int forks)
 
 		/* Forked process */
 		if (pid == 0) {
-			array_clear(*ipc_set);
-			array_push(*ipc_set, sv[0]);
-			close(sv[1]);
 			return forks;
-		/* Parent process */
-		} else {
-			array_push(*ipc_set, sv[1]);
-			/* Do not share parent-end with other forks. */
-			(void) fcntl(sv[1], F_SETFD, FD_CLOEXEC);
-			close(sv[0]);
 		}
 	}
 	return 0;
@@ -234,7 +129,7 @@ static void help(int argc, char *argv[])
 }
 
 /** \return exit code for main()  */
-static int run_worker(uv_loop_t *loop, struct engine *engine, fd_array_t *ipc_set, bool leader, struct args *args)
+static int run_worker(uv_loop_t *loop, struct engine *engine, bool leader, struct args *args)
 {
 	/* Only some kinds of stdin work with uv_pipe_t.
 	 * Otherwise we would abort() from libuv e.g. with </dev/null */
@@ -247,42 +142,37 @@ static int run_worker(uv_loop_t *loop, struct engine *engine, fd_array_t *ipc_se
 	default:
 		kr_log_error(
 			"[system] error: standard input is not a terminal or pipe; "
-			"use '-f 1' if you want non-interactive mode.  "
+			"use '-n' if you want non-interactive mode.  "
 			"Commands can be simply added to your configuration file or sent over the control socket.\n"
 			);
 		return EXIT_FAILURE;
 	}
 
 	/* Control sockets or TTY */
-	uv_pipe_t pipe;
-	uv_pipe_init(loop, &pipe, 0);
-	pipe.data = args;
+	uv_pipe_t *pipe = malloc(sizeof(*pipe));
+	uv_pipe_init(loop, pipe, 0);
 	if (args->interactive) {
 		if (!args->quiet)
 			printf("[system] interactive mode\n> ");
-		uv_pipe_open(&pipe, 0);
-		uv_read_start((uv_stream_t*) &pipe, io_tty_alloc, io_tty_process_input);
-	} else if (args->control_fd != -1 && uv_pipe_open(&pipe, args->control_fd) == 0) {
-		uv_listen((uv_stream_t *) &pipe, 16, io_tty_accept);
+		pipe->data = io_tty_alloc_data();
+		uv_pipe_open(pipe, 0);
+		uv_read_start((uv_stream_t*)pipe, io_tty_alloc, io_tty_process_input);
+	} else if (args->control_fd != -1 && uv_pipe_open(pipe, args->control_fd) == 0) {
+		uv_listen((uv_stream_t *)pipe, 16, io_tty_accept);
 	}
-	/* Watch IPC pipes (or just assign them if leading the pgroup). */
-	if (!leader) {
-		for (size_t i = 0; i < ipc_set->len; ++i) {
-			if (!ipc_watch(loop, engine, ipc_set->at[i])) {
-				kr_log_error("[system] failed to create poller: %s\n", strerror(errno));
-				close(ipc_set->at[i]);
-			}
-		}
-	}
-	memcpy(&engine->ipc_set, ipc_set, sizeof(*ipc_set));
 
 	/* Notify supervisor. */
-#if SYSTEMD_VERSION > 0
+#if ENABLE_LIBSYSTEMD
 	sd_notify(0, "READY=1");
 #endif
 	/* Run event loop */
 	uv_run(loop, UV_RUN_DEFAULT);
-	uv_close((uv_handle_t *)&pipe, NULL); /* Seems OK even on the stopped loop. */
+	/* Free pipe's data.  Seems OK even on the stopped loop.
+	 * In interactive case it may have been done in callbacks already (single leak). */
+	if (!args->interactive) {
+		uv_close((uv_handle_t *)pipe, NULL);
+		free(pipe);
+	}
 	return EXIT_SUCCESS;
 }
 
@@ -490,8 +380,12 @@ static int start_listening(struct network *net, flagged_fd_array_t *fds) {
 /* Drop POSIX 1003.1e capabilities. */
 static void drop_capabilities(void)
 {
-#ifdef ENABLE_CAP_NG
-	/* Drop all capabilities. */
+#if ENABLE_CAP_NG
+	/* Drop all capabilities when running under non-root user. */
+	if (geteuid() == 0) {
+		kr_log_verbose("[system] running as root, no capabilities dropped\n");
+		return;
+	}
 	if (capng_have_capability(CAPNG_EFFECTIVE, CAP_SETPCAP)) {
 		capng_clear(CAPNG_SELECT_BOTH);
 
@@ -499,9 +393,12 @@ static void drop_capabilities(void)
 		if (capng_apply(CAPNG_SELECT_BOTH) < 0) {
 			kr_log_error("[system] failed to set process capabilities: %s\n",
 			          strerror(errno));
+		} else {
+			kr_log_verbose("[system] all capabilities dropped\n");
 		}
 	} else {
-		kr_log_info("[system] process not allowed to set capabilities, skipping\n");
+		/* If user() was called, the capabilities were already dropped along with SETPCAP. */
+		kr_log_verbose("[system] process not allowed to set capabilities, skipping\n");
 	}
 #endif /* ENABLE_CAP_NG */
 }
@@ -515,7 +412,7 @@ int main(int argc, char **argv)
 	}
 	if (strcmp("linux", OPERATING_SYSTEM) != 0)
 		kr_log_info("[warn] Knot Resolver is tested on Linux, other platforms might exhibit bugs.\n"
-				"Please report issues to https://gitlab.labs.nic.cz/knot/knot-resolver/issues/\n"
+				"Please report issues to https://gitlab.nic.cz/knot/knot-resolver/issues/\n"
 				"Thank you for your time and interest!\n");
 
 	the_args = &the_args_value;
@@ -579,11 +476,8 @@ int main(int argc, char **argv)
 				(long)rlim.rlim_cur);
 	}
 
-	/* Connect forks with local socket */
-	fd_array_t ipc_set;
-	array_init(ipc_set);
 	/* Fork subprocesses if requested */
-	int fork_id = fork_workers(&ipc_set, the_args->forks);
+	int fork_id = fork_workers(the_args->forks);
 	if (fork_id < 0) {
 		return EXIT_FAILURE;
 	}
@@ -595,8 +489,6 @@ int main(int argc, char **argv)
 		.ctx = mp_new (4096),
 		.alloc = (knot_mm_alloc_t) mp_alloc
 	};
-	/** Static to work around lua_pushlightuserdata() limitations.
-	 * TODO: convert to a proper singleton like worker, most likely. */
 	static struct engine engine;
 	ret = engine_init(&engine, &pool);
 	if (ret != 0) {
@@ -604,7 +496,7 @@ int main(int argc, char **argv)
 		return EXIT_FAILURE;
 	}
 	/* Initialize the worker */
-	ret = worker_init(&engine, fork_id, the_args->forks);
+	ret = worker_init(&engine, the_args->forks);
 	if (ret != 0) {
 		kr_log_error("[system] failed to initialize worker: %s\n", kr_strerror(ret));
 		return EXIT_FAILURE;
@@ -686,7 +578,7 @@ int main(int argc, char **argv)
 	}
 
 	/* Run the event loop */
-	ret = run_worker(loop, &engine, &ipc_set, fork_id == 0, the_args);
+	ret = run_worker(loop, &engine, fork_id == 0, the_args);
 
 cleanup:/* Cleanup. */
 	engine_deinit(&engine);
