@@ -1,4 +1,4 @@
-/*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2014-2020 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
  *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
@@ -10,6 +10,7 @@
 #include <lauxlib.h>
 #include <libknot/packet/pkt.h>
 #include <libknot/descriptor.h>
+#include <contrib/cleanup.h>
 #include <contrib/ucw/lib.h>
 #include <contrib/ucw/mempool.h>
 #include <contrib/wire.h>
@@ -21,11 +22,16 @@
 #include <unistd.h>
 #include <gnutls/gnutls.h>
 
+#if ENABLE_XDP
+	#include <libknot/xdp/xdp.h>
+#endif
+
 #include "daemon/bindings/api.h"
 #include "daemon/engine.h"
 #include "daemon/io.h"
 #include "daemon/session.h"
 #include "daemon/tls.h"
+#include "daemon/http.h"
 #include "daemon/udp_queue.h"
 #include "daemon/zimport.h"
 #include "lib/layer.h"
@@ -54,15 +60,19 @@ struct request_ctx
 {
 	struct kr_request req;
 
-	struct {
-		/** Requestor's address; separate because of UDP session "sharing". */
-		union inaddr addr;
-		/** NULL if the request didn't come over network. */
-		struct session *session;
-	} source;
-
 	struct worker_ctx *worker;
 	struct qr_task *task;
+	struct {
+		/** NULL if the request didn't come over network. */
+		struct session *session;
+		/** Requestor's address; separate because of UDP session "sharing". */
+		union inaddr addr;
+		/** Local address.  For AF_XDP we couldn't use session's,
+		 * as the address might be different every time. */
+		union inaddr dst_addr;
+		/** MAC addresses - ours [0] and router's [1], in case of AF_XDP socket. */
+		uint8_t eth_addrs[2][6];
+	} source;
 };
 
 /** Query resolution task. */
@@ -118,6 +128,9 @@ static int worker_add_tcp_waiting(struct worker_ctx *worker,
 static struct session* worker_find_tcp_waiting(struct worker_ctx *worker,
 					       const struct sockaddr *addr);
 static void on_tcp_connect_timeout(uv_timer_t *timer);
+static void on_retransmit(uv_timer_t *req);
+static void subreq_finalize(struct qr_task *task, const struct sockaddr *packet_source, knot_pkt_t *pkt);
+
 
 struct worker_ctx the_worker_value; /**< Static allocation is suitable for the singleton. */
 struct worker_ctx *the_worker = NULL;
@@ -125,7 +138,8 @@ struct worker_ctx *the_worker = NULL;
 /*! @internal Create a UDP/TCP handle for an outgoing AF_INET* connection.
  *  socktype is SOCK_* */
 static uv_handle_t *ioreq_spawn(struct worker_ctx *worker,
-				int socktype, sa_family_t family, bool has_tls)
+				int socktype, sa_family_t family, bool has_tls,
+				bool has_http)
 {
 	bool precond = (socktype == SOCK_DGRAM || socktype == SOCK_STREAM)
 			&& (family == AF_INET  || family == AF_INET6);
@@ -141,7 +155,7 @@ static uv_handle_t *ioreq_spawn(struct worker_ctx *worker,
 	if (!handle) {
 		return NULL;
 	}
-	int ret = io_create(worker->loop, handle, socktype, family, has_tls);
+	int ret = io_create(worker->loop, handle, socktype, family, has_tls, has_http);
 	if (ret) {
 		if (ret == UV_EMFILE) {
 			worker->too_many_open = true;
@@ -170,8 +184,7 @@ static uv_handle_t *ioreq_spawn(struct worker_ctx *worker,
 	}
 
 	if (ret != 0) {
-		io_deinit(handle);
-		free(handle);
+		io_free(handle);
 		return NULL;
 	}
 
@@ -252,14 +265,63 @@ static int subreq_key(char *dst, knot_pkt_t *pkt)
 			knot_pkt_qtype(pkt), knot_pkt_qtype(pkt));
 }
 
+#if ENABLE_XDP
+static uint8_t *alloc_wire_cb(struct kr_request *req, uint16_t *maxlen)
+{
+	assert(maxlen);
+	struct request_ctx *ctx = (struct request_ctx *)req;
+	/* We know it's an AF_XDP socket; otherwise this CB isn't assigned. */
+	uv_handle_t *handle = session_get_handle(ctx->source.session);
+	assert(handle->type == UV_POLL);
+	xdp_handle_data_t *xhd = handle->data;
+	knot_xdp_msg_t out;
+	int ret = knot_xdp_send_alloc(xhd->socket, ctx->source.addr.ip.sa_family == AF_INET6,
+					&out, NULL);
+	if (ret != KNOT_EOK) {
+		assert(ret == KNOT_ENOMEM);
+		*maxlen = 0;
+		return NULL;
+	}
+	*maxlen = MIN(*maxlen, out.payload.iov_len);
+	/* It's most convenient to fill the MAC addresses at this point. */
+	memcpy(out.eth_from, &ctx->source.eth_addrs[0], 6);
+	memcpy(out.eth_to,   &ctx->source.eth_addrs[1], 6);
+	return out.payload.iov_base;
+}
+static void free_wire(const struct request_ctx *ctx)
+{
+	assert(ctx->req.alloc_wire_cb == alloc_wire_cb);
+	knot_pkt_t *ans = ctx->req.answer;
+	if (unlikely(ans == NULL)) /* dropped */
+		return;
+	if (likely(ans->wire == NULL)) /* sent most likely */
+		return;
+	/* We know it's an AF_XDP socket; otherwise alloc_wire_cb isn't assigned. */
+	uv_handle_t *handle = session_get_handle(ctx->source.session);
+	assert(handle->type == UV_POLL);
+	xdp_handle_data_t *xhd = handle->data;
+	/* Freeing is done by sending an empty packet (the API won't really send it). */
+	knot_xdp_msg_t out;
+	out.payload.iov_base = ans->wire;
+	out.payload.iov_len = 0;
+	uint32_t sent;
+	int ret = knot_xdp_send(xhd->socket, &out, 1, &sent);
+	assert(ret == KNOT_EOK && sent == 0); (void)ret;
+	kr_log_verbose("[xdp] freed unsent buffer, ret = %d\n", ret);
+}
+#endif
+
 /** Create and initialize a request_ctx (on a fresh mempool).
  *
- * handle and addr point to the source of the request, and they are NULL
+ * session and addr point to the source of the request, and they are NULL
  * in case the request didn't come from network.
  */
 static struct request_ctx *request_create(struct worker_ctx *worker,
 					  struct session *session,
-					  const struct sockaddr *peer,
+					  const struct sockaddr *addr,
+					  const struct sockaddr *dst_addr,
+					  const uint8_t *eth_from,
+					  const uint8_t *eth_to,
 					  uint32_t uid)
 {
 	knot_mm_t pool = {
@@ -282,19 +344,43 @@ static struct request_ctx *request_create(struct worker_ctx *worker,
 		assert(session_flags(session)->outgoing == false);
 	}
 	ctx->source.session = session;
+	assert(!!eth_to == !!eth_from);
+	const bool is_xdp = eth_to != NULL;
+	if (is_xdp) {
+	#if ENABLE_XDP
+		assert(session);
+		memcpy(&ctx->source.eth_addrs[0], eth_to,   sizeof(ctx->source.eth_addrs[0]));
+		memcpy(&ctx->source.eth_addrs[1], eth_from, sizeof(ctx->source.eth_addrs[1]));
+		ctx->req.alloc_wire_cb = alloc_wire_cb;
+	#else
+		assert(!EINVAL);
+		return NULL;
+	#endif
+	}
 
 	struct kr_request *req = &ctx->req;
 	req->pool = pool;
 	req->vars_ref = LUA_NOREF;
 	req->uid = uid;
+	req->qsource.flags.xdp = is_xdp;
 	if (session) {
-		/* We assume the session will be alive during the whole life of the request. */
-		req->qsource.dst_addr = session_get_sockname(session);
 		req->qsource.flags.tcp = session_get_handle(session)->type == UV_TCP;
 		req->qsource.flags.tls = session_flags(session)->has_tls;
+		req->qsource.flags.http = session_flags(session)->has_http;
+		req->qsource.stream_id = -1;
+#if ENABLE_DOH2
+		if (req->qsource.flags.http) {
+			struct http_ctx *http_ctx = session_http_get_server_ctx(session);
+			req->qsource.stream_id = queue_head(http_ctx->streams);
+		}
+#endif
 		/* We need to store a copy of peer address. */
-		memcpy(&ctx->source.addr.ip, peer, kr_sockaddr_len(peer));
+		memcpy(&ctx->source.addr.ip, addr, kr_sockaddr_len(addr));
 		req->qsource.addr = &ctx->source.addr.ip;
+		if (!dst_addr) /* We wouldn't have to copy in this case, but for consistency. */
+			dst_addr = session_get_sockname(session);
+		memcpy(&ctx->source.dst_addr.ip, dst_addr, kr_sockaddr_len(dst_addr));
+		req->qsource.dst_addr = &ctx->source.dst_addr.ip;
 	}
 
 	worker->stats.rconcurrent += 1;
@@ -306,27 +392,13 @@ static struct request_ctx *request_create(struct worker_ctx *worker,
 static int request_start(struct request_ctx *ctx, knot_pkt_t *query)
 {
 	assert(query && ctx);
-	size_t answer_max = KNOT_WIRE_MIN_PKTSIZE;
-	struct kr_request *req = &ctx->req;
 
-	/* source.session can be empty if request was generated by kresd itself */
-	struct session *s = ctx->source.session;
-	if (!s || session_get_handle(s)->type == UV_TCP) {
-		answer_max = KNOT_WIRE_MAX_PKTSIZE;
-	} else if (knot_pkt_has_edns(query)) { /* EDNS */
-		answer_max = MAX(knot_edns_get_payload(query->opt_rr),
-				 KNOT_WIRE_MIN_PKTSIZE);
-	}
+	struct kr_request *req = &ctx->req;
 	req->qsource.size = query->size;
 	if (knot_pkt_has_tsig(query)) {
 		req->qsource.size += query->tsig_wire.len;
 	}
 
-	knot_pkt_t *answer = knot_pkt_new(NULL, answer_max, &req->pool);
-	if (!answer) { /* Failed to allocate answer */
-		return kr_error(ENOMEM);
-	}
-
 	knot_pkt_t *pkt = knot_pkt_new(NULL, req->qsource.size, &req->pool);
 	if (!pkt) {
 		return kr_error(ENOMEM);
@@ -341,7 +413,7 @@ static int request_start(struct request_ctx *ctx, knot_pkt_t *query)
 	/* Start resolution */
 	struct worker_ctx *worker = ctx->worker;
 	struct engine *engine = worker->engine;
-	kr_resolve_begin(req, &engine->resolver, answer);
+	kr_resolve_begin(req, &engine->resolver);
 	worker->stats.queries += 1;
 	/* Throttle outbound queries only when high pressure */
 	if (worker->stats.concurrent < QUERY_RATE_THRESHOLD) {
@@ -367,6 +439,14 @@ static void request_free(struct request_ctx *ctx)
 		lua_pop(L, 1);
 		ctx->req.vars_ref = LUA_NOREF;
 	}
+	/* Make sure to free XDP buffer in case it wasn't sent. */
+	if (ctx->req.alloc_wire_cb) {
+	#if ENABLE_XDP
+		free_wire(ctx);
+	#else
+		assert(!EINVAL);
+	#endif
+	}
 	/* Return mempool to ring or free it if it's full */
 	pool_release(worker, ctx->req.pool.ctx);
 	/* @note The 'task' is invalidated from now on. */
@@ -377,9 +457,9 @@ static struct qr_task *qr_task_create(struct request_ctx *ctx)
 {
 	/* Choose (initial) pktbuf size.  As it is now, pktbuf can be used
 	 * for UDP answers from upstream *and* from cache
-	 * and for sending non-UDP queries upstream (?) */
+	 * and for sending queries upstream */
 	uint16_t pktbuf_max = KR_EDNS_PAYLOAD;
-	const knot_rrset_t *opt_our = ctx->worker->engine->resolver.opt_rr;
+	const knot_rrset_t *opt_our = ctx->worker->engine->resolver.upstream_opt_rr;
 	if (opt_our) {
 		pktbuf_max = MAX(pktbuf_max, knot_edns_get_payload(opt_our));
 	}
@@ -477,7 +557,7 @@ static void qr_task_complete(struct qr_task *task)
 }
 
 /* This is called when we send subrequest / answer */
-int qr_task_on_send(struct qr_task *task, uv_handle_t *handle, int status)
+int qr_task_on_send(struct qr_task *task, const uv_handle_t *handle, int status)
 {
 
 	if (task->finished) {
@@ -485,27 +565,52 @@ int qr_task_on_send(struct qr_task *task, uv_handle_t *handle, int status)
 		qr_task_complete(task);
 	}
 
-	if (!handle || handle->type != UV_TCP) {
+	if (!handle)
 		return status;
-	}
 
 	struct session* s = handle->data;
 	assert(s);
-	if (status != 0) {
-		session_tasklist_del(s, task);
+
+	if (handle->type == UV_UDP && session_flags(s)->outgoing) {
+		/* Start the timeout timer for UDP here, since this is the closest
+		 * to the wire we can get. */
+		struct kr_request *req = &task->ctx->req;
+		/* Check current query NSLIST */
+		struct kr_query *qry = array_tail(req->rplan.pending);
+		assert(qry != NULL);
+		/* Retransmit at default interval, or more frequently if the mean
+		 * RTT of the server is better. If the server is glued, use default rate. */
+		size_t timeout = qry->ns.score;
+		if (timeout > KR_NS_GLUED) {
+			/* We don't have information about variance in RTT, expect +10ms */
+			timeout = MIN(qry->ns.score + 10, KR_CONN_RETRY);
+		} else {
+			timeout = KR_CONN_RETRY;
+		}
+
+		int ret = session_timer_start(s, on_retransmit, timeout, 0);
+		/* Start next step with timeout, fatal if can't start a timer. */
+		if (ret != 0) {
+			subreq_finalize(task, &qry->ns.addr->ip, task->pktbuf);
+			qr_task_finalize(task, KR_STATE_FAIL);
+		}
 	}
 
-	if (session_flags(s)->outgoing || session_flags(s)->closing) {
-		return status;
-	}
+	if (handle->type == UV_TCP) {
+		if (status != 0)
+			session_tasklist_del(s, task);
 
-	struct worker_ctx *worker = task->ctx->worker;
-	if (session_flags(s)->throttled &&
-	    session_tasklist_get_len(s) < worker->tcp_pipeline_max/2) {
-	   /* Start reading again if the session is throttled and
-	    * the number of outgoing requests is below watermark. */
-		session_start_read(s);
-		session_flags(s)->throttled = false;
+		if (session_flags(s)->outgoing || session_flags(s)->closing)
+			return status;
+
+		struct worker_ctx *worker = task->ctx->worker;
+		if (session_flags(s)->throttled &&
+		    session_tasklist_get_len(s) < worker->tcp_pipeline_max/2) {
+			/* Start reading again if the session is throttled and
+			 * the number of outgoing requests is below watermark. */
+			session_start_read(s);
+			session_flags(s)->throttled = false;
+		}
 	}
 
 	return status;
@@ -578,7 +683,15 @@ static int qr_task_send(struct qr_task *task, struct session *session,
 	struct worker_ctx *worker = ctx->worker;
 	/* Send using given protocol */
 	assert(!session_flags(session)->closing);
-	if (session_flags(session)->has_tls) {
+	if (session_flags(session)->has_http) {
+#if ENABLE_DOH2
+		uv_write_t *write_req = (uv_write_t *)ioreq;
+		write_req->data = task;
+		ret = http_write(write_req, handle, pkt, ctx->req.qsource.stream_id, &on_write);
+#else
+		ret = kr_error(ENOPROTOOPT);
+#endif
+	} else if (session_flags(session)->has_tls) {
 		uv_write_t *write_req = (uv_write_t *)ioreq;
 		write_req->data = task;
 		ret = tls_write(write_req, handle, pkt, &on_write);
@@ -633,9 +746,18 @@ static int qr_task_send(struct qr_task *task, struct session *session,
 			worker->rconcurrent_highwatermark = worker->stats.rconcurrent;
 			ret = kr_error(UV_EMFILE);
 		}
+
+		if (session_flags(session)->has_http)
+			worker->stats.err_http += 1;
+		else if (session_flags(session)->has_tls)
+			worker->stats.err_tls += 1;
+		else if (handle->type == UV_UDP)
+			worker->stats.err_udp += 1;
+		else
+			worker->stats.err_tcp += 1;
 	}
 
-	/* Update statistics */
+	/* Update outgoing query statistics */
 	if (session_flags(session)->outgoing && addr) {
 		if (session_flags(session)->has_tls)
 			worker->stats.tls += 1;
@@ -664,11 +786,8 @@ static struct kr_query *task_get_last_pending_query(struct qr_task *task)
 static int session_tls_hs_cb(struct session *session, int status)
 {
 	assert(session_flags(session)->outgoing);
-	uv_handle_t *handle = session_get_handle(session);
-	uv_loop_t *loop = handle->loop;
-	struct worker_ctx *worker = loop->data;
 	struct sockaddr *peer = session_get_peer(session);
-	int deletion_res = worker_del_tcp_waiting(worker, peer);
+	int deletion_res = worker_del_tcp_waiting(the_worker, peer);
 	int ret = kr_ok();
 
 	if (status) {
@@ -677,7 +796,7 @@ static int session_tls_hs_cb(struct session *session, int status)
 			struct kr_qflags *options = &task->ctx->req.options;
 			unsigned score = options->FORWARD || options->STUB ? KR_NS_FWD_DEAD : KR_NS_DEAD;
 			kr_nsrep_update_rtt(NULL, peer, score,
-					    worker->engine->resolver.cache_rtt,
+					    the_worker->engine->resolver.cache_rtt,
 					    KR_NS_UPDATE_NORESET);
 		}
 #ifndef NDEBUG
@@ -689,14 +808,14 @@ static int session_tls_hs_cb(struct session *session, int status)
 			assert(deletion_res != 0);
 			const char *key = tcpsess_key(peer);
 			assert(key);
-			assert(map_contains(&worker->tcp_connected, key) != 0);
+			assert(map_contains(&the_worker->tcp_connected, key) != 0);
 		}
 #endif
 		return ret;
 	}
 
 	/* handshake was completed successfully */
-	struct tls_client_ctx_t *tls_client_ctx = session_tls_get_client_ctx(session);
+	struct tls_client_ctx *tls_client_ctx = session_tls_get_client_ctx(session);
 	tls_client_param_t *tls_params = tls_client_ctx->params;
 	gnutls_session_t tls_session = tls_client_ctx->c.tls_session;
 	if (gnutls_session_is_resumed(tls_session) != 0) {
@@ -717,7 +836,7 @@ static int session_tls_hs_cb(struct session *session, int status)
 		}
 	}
 
-	struct session *s = worker_find_tcp_connected(worker, peer);
+	struct session *s = worker_find_tcp_connected(the_worker, peer);
 	ret = kr_ok();
 	if (deletion_res == kr_ok()) {
 		/* peer was in the waiting list, add to the connected list. */
@@ -726,7 +845,7 @@ static int session_tls_hs_cb(struct session *session, int status)
 			 * peer already is in the connected list. */
 			ret = kr_error(EINVAL);
 		} else {
-			ret = worker_add_tcp_connected(worker, peer, session);
+			ret = worker_add_tcp_connected(the_worker, peer, session);
 		}
 	} else {
 		/* peer wasn't in the waiting list.
@@ -758,9 +877,9 @@ static int session_tls_hs_cb(struct session *session, int status)
 		/* Something went wrong.
 		 * Either addition to the list of connected sessions
 		 * or write to upstream failed. */
-		worker_del_tcp_connected(worker, peer);
+		worker_del_tcp_connected(the_worker, peer);
 		session_waitinglist_finalize(session, KR_STATE_FAIL);
-		assert(session_tasklist_is_empty(session));
+		session_tasklist_finalize(session, KR_STATE_FAIL);
 		session_close(session);
 	} else {
 		session_timer_stop(session);
@@ -891,7 +1010,7 @@ static void on_connect(uv_connect_t *req, int status)
 
 	int ret = kr_ok();
 	if (session_flags(session)->has_tls) {
-		struct tls_client_ctx_t *tls_ctx = session_tls_get_client_ctx(session);
+		struct tls_client_ctx *tls_ctx = session_tls_get_client_ctx(session);
 		ret = tls_client_connect_start(tls_ctx, session, session_tls_hs_cb);
 		if (ret == kr_error(EAGAIN)) {
 			session_timer_stop(session);
@@ -1009,7 +1128,7 @@ static uv_handle_t *retransmit(struct qr_task *task)
 		if (kr_resolve_checkout(&ctx->req, NULL, (struct sockaddr *)choice, SOCK_DGRAM, task->pktbuf) != 0) {
 			return ret;
 		}
-		ret = ioreq_spawn(ctx->worker, SOCK_DGRAM, choice->sin6_family, false);
+		ret = ioreq_spawn(ctx->worker, SOCK_DGRAM, choice->sin6_family, false, false);
 		if (!ret) {
 			return ret;
 		}
@@ -1026,7 +1145,7 @@ static uv_handle_t *retransmit(struct qr_task *task)
 			task->pending[task->pending_count] = session;
 			task->pending_count += 1;
 			task->addrlist_turn = (task->addrlist_turn + 1) %
-					      task->addrlist_count; /* Round robin */
+					       task->addrlist_count; /* Round robin */
 			session_start_read(session); /* Start reading answer */
 		}
 	}
@@ -1125,11 +1244,55 @@ static bool subreq_enqueue(struct qr_task *task)
 	return true;
 }
 
+#if ENABLE_XDP
+static void xdp_tx_waker(uv_idle_t *handle)
+{
+	int ret = knot_xdp_send_finish(handle->data);
+	if (ret != KNOT_EAGAIN && ret != KNOT_EOK)
+		kr_log_error("[xdp] check: ret = %d, %s\n", ret, knot_strerror(ret));
+	/* Apparently some drivers need many explicit wake-up calls
+	 * even if we push no additional packets (in case they accumulated a lot) */
+	if (ret != KNOT_EAGAIN)
+		uv_idle_stop(handle);
+	knot_xdp_send_prepare(handle->data);
+	/* LATER(opt.): it _might_ be better for performance to do these two steps
+	 * at different points in time */
+}
+#endif
+/** Send an answer packet over XDP. */
+static int xdp_push(struct qr_task *task, const uv_handle_t *src_handle)
+{
+#if ENABLE_XDP
+	struct request_ctx *ctx = task->ctx;
+	knot_xdp_msg_t msg;
+	const struct sockaddr *ip_from = &ctx->source.dst_addr.ip;
+	const struct sockaddr *ip_to   = &ctx->source.addr.ip;
+	memcpy(&msg.ip_from, ip_from, kr_sockaddr_len(ip_from));
+	memcpy(&msg.ip_to,   ip_to,   kr_sockaddr_len(ip_to));
+	msg.payload.iov_base = ctx->req.answer->wire;
+	msg.payload.iov_len  = ctx->req.answer->size;
+
+	xdp_handle_data_t *xhd = src_handle->data;
+	assert(xhd && xhd->socket && xhd->session == ctx->source.session);
+	uint32_t sent;
+	int ret = knot_xdp_send(xhd->socket, &msg, 1, &sent);
+	ctx->req.answer->wire = NULL; /* it's been freed */
+
+	uv_idle_start(&xhd->tx_waker, xdp_tx_waker);
+	kr_log_verbose("[xdp] pushed a packet, ret = %d\n", ret);
+
+	return qr_task_on_send(task, src_handle, ret);
+#else
+	assert(!EINVAL);
+	return kr_error(EINVAL);
+#endif
+}
+
 static int qr_task_finalize(struct qr_task *task, int state)
 {
 	assert(task && task->leading == false);
 	if (task->finished) {
-		return 0;
+		return kr_ok();
 	}
 	struct request_ctx *ctx = task->ctx;
 	struct session *source_session = ctx->source.session;
@@ -1138,21 +1301,32 @@ static int qr_task_finalize(struct qr_task *task, int state)
 	task->finished = true;
 	if (source_session == NULL) {
 		(void) qr_task_on_send(task, NULL, kr_error(EIO));
-		return state == KR_STATE_DONE ? 0 : kr_error(EIO);
+		return state == KR_STATE_DONE ? kr_ok() : kr_error(EIO);
 	}
 
+	if (unlikely(ctx->req.answer == NULL)) { /* meant to be dropped */
+		(void) qr_task_on_send(task, NULL, kr_ok());
+		return kr_ok();
+	}
+
+	if (session_flags(source_session)->closing ||
+	    ctx->source.addr.ip.sa_family == AF_UNSPEC)
+		return kr_error(EINVAL);
+
 	/* Reference task as the callback handler can close it */
 	qr_task_ref(task);
 
 	/* Send back answer */
-	assert(!session_flags(source_session)->closing);
-	assert(ctx->source.addr.ip.sa_family != AF_UNSPEC);
-
 	int ret;
 	const uv_handle_t *src_handle = session_get_handle(source_session);
-	if (src_handle->type != UV_UDP && src_handle->type != UV_TCP) {
+	if (src_handle->type != UV_UDP && src_handle->type != UV_TCP
+				       && src_handle->type != UV_POLL) {
 		assert(false);
 		ret = kr_error(EINVAL);
+
+	} else if (src_handle->type == UV_POLL) {
+		ret = xdp_push(task, src_handle);
+
 	} else if (src_handle->type == UV_UDP && ENABLE_SENDMMSG) {
 		int fd;
 		ret = uv_fileno(src_handle, &fd);
@@ -1182,15 +1356,14 @@ static int qr_task_finalize(struct qr_task *task, int state)
 
 	qr_task_unref(task);
 
-	return state == KR_STATE_DONE ? 0 : kr_error(EIO);
+	if (ret != kr_ok() || state != KR_STATE_DONE)
+		return kr_error(EIO);
+	return kr_ok();
 }
 
 static int udp_task_step(struct qr_task *task,
 			 const struct sockaddr *packet_source, knot_pkt_t *packet)
 {
-	struct request_ctx *ctx = task->ctx;
-	struct kr_request *req = &ctx->req;
-
 	/* If there is already outgoing query, enqueue to it. */
 	if (subreq_enqueue(task)) {
 		return kr_ok(); /* Will be notified when outgoing query finishes. */
@@ -1201,30 +1374,12 @@ static int udp_task_step(struct qr_task *task,
 		subreq_finalize(task, packet_source, packet);
 		return qr_task_finalize(task, KR_STATE_FAIL);
 	}
-	/* Check current query NSLIST */
-	struct kr_query *qry = array_tail(req->rplan.pending);
-	assert(qry != NULL);
-	/* Retransmit at default interval, or more frequently if the mean
-	 * RTT of the server is better. If the server is glued, use default rate. */
-	size_t timeout = qry->ns.score;
-	if (timeout > KR_NS_GLUED) {
-		/* We don't have information about variance in RTT, expect +10ms */
-		timeout = MIN(qry->ns.score + 10, KR_CONN_RETRY);
-	} else {
-		timeout = KR_CONN_RETRY;
-	}
+
 	/* Announce and start subrequest.
 	 * @note Only UDP can lead I/O as it doesn't touch 'task->pktbuf' for reassembly.
 	 */
 	subreq_lead(task);
-	struct session *session = handle->data;
-	assert(session_get_handle(session) == handle && (handle->type == UV_UDP));
-	int ret = session_timer_start(session, on_retransmit, timeout, 0);
-	/* Start next step with timeout, fatal if can't start a timer. */
-	if (ret != 0) {
-		subreq_finalize(task, packet_source, packet);
-		return qr_task_finalize(task, KR_STATE_FAIL);
-	}
+
 	return kr_ok();
 }
 
@@ -1293,7 +1448,7 @@ static int tcp_task_make_connection(struct qr_task *task, const struct sockaddr
 	struct worker_ctx *worker = ctx->worker;
 
 	/* Check if there must be TLS */
-	struct tls_client_ctx_t *tls_ctx = NULL;
+	struct tls_client_ctx *tls_ctx = NULL;
 	struct network *net = &worker->engine->net;
 	tls_client_param_t *entry = tls_client_param_get(net->tls_client_params, addr);
 	if (entry) {
@@ -1310,8 +1465,9 @@ static int tcp_task_make_connection(struct qr_task *task, const struct sockaddr
 		tls_client_ctx_free(tls_ctx);
 		return kr_error(EINVAL);
 	}
+	bool has_http = false;
 	bool has_tls = (tls_ctx != NULL);
-	uv_handle_t *client = ioreq_spawn(worker, SOCK_STREAM, addr->sa_family, has_tls);
+	uv_handle_t *client = ioreq_spawn(worker, SOCK_STREAM, addr->sa_family, has_tls, has_http);
 	if (!client) {
 		tls_client_ctx_free(tls_ctx);
 		free(conn);
@@ -1562,32 +1718,38 @@ static int parse_packet(knot_pkt_t *query)
 	return ret;
 }
 
-int worker_submit(struct session *session, const struct sockaddr *peer, knot_pkt_t *query)
+int worker_submit(struct session *session,
+		  const struct sockaddr *peer, const struct sockaddr *dst_addr,
+		  const uint8_t *eth_from, const uint8_t *eth_to, knot_pkt_t *pkt)
 {
-	if (!session) {
-		assert(false);
+	if (!session || !pkt)
 		return kr_error(EINVAL);
-	}
 
 	uv_handle_t *handle = session_get_handle(session);
-	bool OK = handle && handle->loop->data;
-	if (!OK) {
-		assert(false);
+	if (!handle || !handle->loop->data)
 		return kr_error(EINVAL);
-	}
 
-	struct worker_ctx *worker = handle->loop->data;
+	int ret = parse_packet(pkt);
 
-	/* Parse packet */
-	int ret = parse_packet(query);
-
-	const bool is_query = (knot_wire_get_qr(query->wire) == 0);
+	const bool is_query = (knot_wire_get_qr(pkt->wire) == 0);
 	const bool is_outgoing = session_flags(session)->outgoing;
+
+	struct http_ctx *http_ctx = NULL;
+#if ENABLE_DOH2
+	http_ctx = session_http_get_server_ctx(session);
+#endif
+
+	if (!is_outgoing && http_ctx && queue_len(http_ctx->streams) <= 0)
+		return kr_error(ENOENT);
+
 	/* Ignore badly formed queries. */
-	if (!query ||
-	    (ret != kr_ok() && ret != kr_error(EMSGSIZE)) ||
+	if ((ret != kr_ok() && ret != kr_error(EMSGSIZE)) ||
 	    (is_query == is_outgoing)) {
-		if (query && !is_outgoing) worker->stats.dropped += 1;
+		if (!is_outgoing) {
+			the_worker->stats.dropped += 1;
+			if (http_ctx)
+				queue_pop(http_ctx->streams);
+		}
 		return kr_error(EILSEQ);
 	}
 
@@ -1596,13 +1758,15 @@ int worker_submit(struct session *session, const struct sockaddr *peer, knot_pkt
 	struct qr_task *task = NULL;
 	const struct sockaddr *addr = NULL;
 	if (!is_outgoing) { /* request from a client */
-		struct request_ctx *ctx = request_create(worker, session, peer,
-							 knot_wire_get_id(query->wire));
-		if (!ctx) {
+		struct request_ctx *ctx =
+			request_create(the_worker, session, peer, dst_addr,
+					eth_from, eth_to, knot_wire_get_id(pkt->wire));
+		if (http_ctx)
+			queue_pop(http_ctx->streams);
+		if (!ctx)
 			return kr_error(ENOMEM);
-		}
 
-		ret = request_start(ctx, query);
+		ret = request_start(ctx, pkt);
 		if (ret != 0) {
 			request_free(ctx);
 			return kr_error(ENOMEM);
@@ -1617,8 +1781,8 @@ int worker_submit(struct session *session, const struct sockaddr *peer, knot_pkt
 		if (handle->type == UV_TCP && qr_task_register(task, session)) {
 			return kr_error(ENOMEM);
 		}
-	} else if (query) { /* response from upstream */
-		const uint16_t id = knot_wire_get_id(query->wire);
+	} else { /* response from upstream */
+		const uint16_t id = knot_wire_get_id(pkt->wire);
 		task = session_tasklist_del_msgid(session, id);
 		if (task == NULL) {
 			VERBOSE_MSG(NULL, "=> ignoring packet with mismatching ID %d\n",
@@ -1634,7 +1798,7 @@ int worker_submit(struct session *session, const struct sockaddr *peer, knot_pkt
 	 * Task was created (found). */
 	session_touch(session);
 	/* Consume input and produce next message */
-	return qr_task_step(task, addr, query);
+	return qr_task_step(task, addr, pkt);
 }
 
 static int map_add_tcp_session(map_t *map, const struct sockaddr* addr,
@@ -1727,21 +1891,19 @@ int worker_end_tcp(struct session *session)
 
 	session_timer_stop(session);
 
-	uv_handle_t *handle = session_get_handle(session);
-	struct worker_ctx *worker = handle->loop->data;
 	struct sockaddr *peer = session_get_peer(session);
 
-	worker_del_tcp_waiting(worker, peer);
-	worker_del_tcp_connected(worker, peer);
+	worker_del_tcp_waiting(the_worker, peer);
+	worker_del_tcp_connected(the_worker, peer);
 	session_flags(session)->connected = false;
 
-	struct tls_client_ctx_t *tls_client_ctx = session_tls_get_client_ctx(session);
+	struct tls_client_ctx *tls_client_ctx = session_tls_get_client_ctx(session);
 	if (tls_client_ctx) {
 		/* Avoid gnutls_bye() call */
 		tls_set_hs_state(&tls_client_ctx->c, TLS_HS_NOT_STARTED);
 	}
 
-	struct tls_ctx_t *tls_ctx = session_tls_get_server_ctx(session);
+	struct tls_ctx *tls_ctx = session_tls_get_server_ctx(session);
 	if (tls_ctx) {
 		/* Avoid gnutls_bye() call */
 		tls_set_hs_state(&tls_ctx->c, TLS_HS_NOT_STARTED);
@@ -1804,7 +1966,7 @@ knot_pkt_t * worker_resolve_mk_pkt(const char *qname_str, uint16_t qtype, uint16
 
 	/* Add OPT RR, including wire format so modules can see both representations.
 	 * knot_pkt_put() copies the outside; we need to duplicate the inside manually. */
-	knot_rrset_t *opt = knot_rrset_copy(the_worker->engine->resolver.opt_rr, NULL);
+	knot_rrset_t *opt = knot_rrset_copy(the_worker->engine->resolver.downstream_opt_rr, NULL);
 	if (!opt) {
 		knot_pkt_free(pkt);
 		return NULL;
@@ -1838,7 +2000,8 @@ struct qr_task *worker_resolve_start(knot_pkt_t *query, struct kr_qflags options
 	}
 
 
-	struct request_ctx *ctx = request_create(worker, NULL, NULL, worker->next_request_uid);
+	struct request_ctx *ctx = request_create(worker, NULL, NULL, NULL, NULL, NULL,
+						 worker->next_request_uid);
 	if (!ctx) {
 		return NULL;
 	}
@@ -2022,7 +2185,7 @@ void worker_deinit(void)
 	the_worker = NULL;
 }
 
-int worker_init(struct engine *engine, int worker_id, int worker_count)
+int worker_init(struct engine *engine, int worker_count)
 {
 	assert(engine && engine->L);
 	assert(the_worker == NULL);
@@ -2036,7 +2199,6 @@ int worker_init(struct engine *engine, int worker_id, int worker_count)
 	uv_loop_t *loop = uv_default_loop();
 	worker->loop = loop;
 
-	worker->id = worker_id;
 	worker->count = worker_count;
 
 	/* Register table for worker per-request variables */
@@ -2056,9 +2218,20 @@ int worker_init(struct engine *engine, int worker_id, int worker_count)
 
 	/* Set some worker.* fields in Lua */
 	lua_getglobal(engine->L, "worker");
-	lua_pushnumber(engine->L, worker_id);
+	pid_t pid = getpid();
+
+	auto_free char *pid_str = NULL;
+	const char *inst_name = getenv("SYSTEMD_INSTANCE");
+	if (inst_name) {
+		lua_pushstring(engine->L, inst_name);
+	} else {
+		ret = asprintf(&pid_str, "%ld", (long)pid);
+		assert(ret > 0);
+		lua_pushstring(engine->L, pid_str);
+	}
 	lua_setfield(engine->L, -2, "id");
-	lua_pushnumber(engine->L, getpid());
+
+	lua_pushnumber(engine->L, pid);
 	lua_setfield(engine->L, -2, "pid");
 	lua_pushnumber(engine->L, worker_count);
 	lua_setfield(engine->L, -2, "count");
@@ -2070,7 +2243,7 @@ int worker_init(struct engine *engine, int worker_id, int worker_count)
 
 	the_worker = worker;
 	loop->data = the_worker;
-	/* ^^^^ This shouldn't be used anymore, but it's hard to be 100% sure. */
+	/* ^^^^ Now this shouldn't be used anymore, but it's hard to be 100% sure. */
 	return kr_ok();
 }
 
