@@ -23,7 +23,7 @@ KR_EXPORT extern struct worker_ctx *the_worker;
 
 /** Create and initialize the worker.
  * \return error code (ENOMEM) */
-int worker_init(struct engine *engine, int worker_id, int worker_count);
+int worker_init(struct engine *engine, int worker_count);
 
 /** Destroy the worker (free memory). */
 void worker_deinit(void);
@@ -31,12 +31,15 @@ void worker_deinit(void);
 /**
  * Process an incoming packet (query from a client or answer from upstream).
  *
- * @param session  session the packet came from
- * @param peer     address the packet came from
- * @param query    the packet, or NULL on an error from the transport layer
+ * @param session  session the packet came from, or NULL (not from network)
+ * @param peer     address the packet came from, or NULL (not from network)
+ * @param eth_*    MAC addresses or NULL (they're useful for XDP)
+ * @param pkt      the packet, or NULL (an error from the transport layer)
  * @return 0 or an error code
  */
-int worker_submit(struct session *session, const struct sockaddr *peer, knot_pkt_t *query);
+int worker_submit(struct session *session,
+		  const struct sockaddr *peer, const struct sockaddr *dst_addr,
+		  const uint8_t *eth_from, const uint8_t *eth_to, knot_pkt_t *pkt);
 
 /**
  * End current DNS/TCP session, this disassociates pending tasks from this session
@@ -108,7 +111,7 @@ void worker_task_subreq_finalize(struct qr_task *task);
 bool worker_task_finished(struct qr_task *task);
 
 /** To be called after sending a DNS message.  It mainly deals with cleanups. */
-int qr_task_on_send(struct qr_task *task, uv_handle_t *handle, int status);
+int qr_task_on_send(struct qr_task *task, const uv_handle_t *handle, int status);
 
 /** Various worker statistics.  Sync with wrk_stats() */
 struct worker_stats {
@@ -123,6 +126,11 @@ struct worker_stats {
 	size_t tls;  /**< Number of outbound queries over TLS. */
 	size_t ipv4; /**< Number of outbound queries over IPv4.*/
 	size_t ipv6; /**< Number of outbound queries over IPv6. */
+
+	size_t err_udp;  /**< Total number of write errors for UDP transport. */
+	size_t err_tcp;  /**< Total number of write errors for TCP transport. */
+	size_t err_tls;  /**< Total number of write errors for TLS transport. */
+	size_t err_http;  /**< Total number of write errors for HTTP(S) transport. */
 };
 
 /** @cond internal */
@@ -147,8 +155,7 @@ typedef array_t(struct qr_task *) qr_tasklist_t;
 struct worker_ctx {
 	struct engine *engine;
 	uv_loop_t *loop;
-	int id;
-	int count;
+	int count;  /** unreliable, does not count systemd instance, do not use */
 	int vars_table_ref;
 	unsigned tcp_pipeline_max;
 
