@@ -15,6 +15,8 @@
 #include <libknot/errcode.h>
 #include <libknot/rrtype/rrsig.h>
 
+#include <uv.h>
+
 #include "contrib/base32hex.h"
 #include "contrib/cleanup.h"
 #include "contrib/ucw/lib.h"
@@ -61,7 +63,7 @@ static ssize_t stash_rrset(struct kr_cache *cache, const struct kr_query *qry,
 /** Preliminary checks before stash_rrset().  Don't call if returns <= 0. */
 static int stash_rrset_precond(const knot_rrset_t *rr, const struct kr_query *qry/*logs*/);
 
-/** @internal Open cache db transaction and check internal data version. */
+/** @internal Ensure the cache version is right, possibly by clearing it. */
 static int assert_right_version(struct kr_cache *cache)
 {
 	/* Check cache ABI version. */
@@ -72,31 +74,34 @@ static int assert_right_version(struct kr_cache *cache)
 	int ret = cache_op(cache, read, &key, &val, 1);
 	if (ret == 0 && val.len == sizeof(CACHE_VERSION)
 	    && memcmp(val.data, &CACHE_VERSION, sizeof(CACHE_VERSION)) == 0) {
-		ret = kr_error(EEXIST);
+		ret = kr_ok();
 	} else {
 		int oldret = ret;
-		/* Version doesn't match. Recreate cache and write version key. */
+		/* Version doesn't match or we were unable to read it, possibly because DB is empty.
+		 * Recreate cache and write version key. */
 		ret = cache_op(cache, count);
-		if (ret != 0) { /* Non-empty cache, purge it. */
-			kr_log_info("[     ][cach] incompatible cache database detected, purging\n");
+		if (ret != 0) { /* Log for non-empty cache to limit noise on fresh start. */
+			kr_log_info("[cache] incompatible cache database detected, purging\n");
 			if (oldret) {
-				kr_log_verbose("bad ret: %d\n", oldret);
+				kr_log_verbose("[cache] reading version returned: %d\n", oldret);
 			} else if (val.len != sizeof(CACHE_VERSION)) {
-				kr_log_verbose("bad length: %d\n", (int)val.len);
+				kr_log_verbose("[cache] version has bad length: %d\n", (int)val.len);
 			} else {
 				uint16_t ver;
 				memcpy(&ver, val.data, sizeof(ver));
-				kr_log_verbose("bad version: %d\n", (int)ver);
+				kr_log_verbose("[cache] version has bad value: %d instead of %d\n",
+					(int)ver, (int)CACHE_VERSION);
 			}
-			ret = cache_op(cache, clear);
-		}
-		/* Either purged or empty. */
-		if (ret == 0) {
-			/* Key/Val is invalidated by cache purge, recreate it */
-			val.data = /*const-cast*/(void *)&CACHE_VERSION;
-			val.len = sizeof(CACHE_VERSION);
-			ret = cache_op(cache, write, &key, &val, 1);
 		}
+		ret = cache_op(cache, clear);
+	}
+	/* Rewrite the entry even if it isn't needed.  Because of cache-size-changing
+	 * possibility it's good to always perform some write during opening of cache. */
+	if (ret == 0) {
+		/* Key/Val is invalidated by cache purge, recreate it */
+		val.data = /*const-cast*/(void *)&CACHE_VERSION;
+		val.len = sizeof(CACHE_VERSION);
+		ret = cache_op(cache, write, &key, &val, 1);
 	}
 	kr_cache_commit(cache);
 	return ret;
@@ -105,31 +110,52 @@ static int assert_right_version(struct kr_cache *cache)
 int kr_cache_open(struct kr_cache *cache, const struct kr_cdb_api *api, struct kr_cdb_opts *opts, knot_mm_t *mm)
 {
 	if (!cache) {
+		assert(cache);
 		return kr_error(EINVAL);
 	}
+	memset(cache, 0, sizeof(*cache));
 	/* Open cache */
 	if (!api) {
 		api = kr_cdb_lmdb();
 	}
 	cache->api = api;
-	memset(&cache->stats, 0, sizeof(cache->stats));
 	int ret = cache->api->open(&cache->db, &cache->stats, opts, mm);
+	if (ret == 0) {
+		ret = assert_right_version(cache);
+		// The included write also committed maxsize increase to the file.
+	}
+	if (ret == 0 && opts->maxsize) {
+		/* If some maxsize is requested and it's smaller than in-file maxsize,
+		 * LMDB only restricts our env without changing the in-file maxsize.
+		 * That is worked around by reopening (found no other reliable way). */
+		cache->api->close(cache->db, &cache->stats);
+		struct kr_cdb_opts opts2;
+		memcpy(&opts2, opts, sizeof(opts2));
+		opts2.maxsize = 0;
+		ret = cache->api->open(&cache->db, &cache->stats, &opts2, mm);
+	}
+
+	char *fpath = kr_absolutize_path(opts->path, "data.mdb");
+	if (fpath) {
+		kr_cache_emergency_file_to_remove = fpath;
+	} else {
+		assert(false); /* non-critical, but still */
+		fpath = "<ENOMEM>";
+	}
+
+	if (ret == 0 && opts->maxsize) {
+		size_t maxsize = cache->api->get_maxsize(cache->db);
+		if (maxsize > opts->maxsize) kr_log_info(
+			"[cache] Warning: real cache size is %zu instead of the requested %zu bytes."
+			"  To reduce the size you need to remove the file '%s' by hand.\n",
+			maxsize, opts->maxsize, fpath);
+	}
 	if (ret != 0) {
 		return ret;
 	}
 	cache->ttl_min = KR_CACHE_DEFAULT_TTL_MIN;
 	cache->ttl_max = KR_CACHE_DEFAULT_TTL_MAX;
-	/* Check cache ABI version */
 	kr_cache_make_checkpoint(cache);
-	(void)assert_right_version(cache);
-
-	char *fpath;
-	ret = asprintf(&fpath, "%s/data.mdb", opts->path);
-	if (ret > 0) {
-		kr_cache_emergency_file_to_remove = fpath;
-	} else {
-		assert(false); /* non-critical, but still */
-	}
 	return 0;
 }
 
@@ -140,6 +166,7 @@ const char *kr_cache_emergency_file_to_remove = NULL;
 
 void kr_cache_close(struct kr_cache *cache)
 {
+	kr_cache_check_health(cache, -1);
 	if (cache_isvalid(cache)) {
 		cache_op(cache, close);
 		cache->db = NULL;
@@ -941,3 +968,46 @@ cleanup:
 	return ret;
 }
 
+static void health_timer_cb(uv_timer_t *health_timer)
+{
+	struct kr_cache *cache = health_timer->data;
+	if (cache)
+		cache_op(cache, check_health);
+	/* We don't do anything with the return code.  For example, in some situations
+	 * the file may not exist (temporarily), and we just expect to be more lucky
+	 * when the timer fires again. */
+}
+
+int kr_cache_check_health(struct kr_cache *cache, int interval)
+{
+	if (interval == 0) {
+		return cache_op(cache, check_health);
+	}
+	if (interval < 0) {
+		if (!cache->health_timer)
+			return kr_ok(); // tolerate stopping a "stopped" timer
+		uv_close((uv_handle_t *)cache->health_timer, (uv_close_cb)free);
+		cache->health_timer->data = NULL;
+		cache->health_timer = NULL;
+		return kr_ok();
+	}
+
+	assert(interval > 0);
+	if (!cache->health_timer) {
+		/* We avoid depending on daemon's symbols by using uv_default_loop. */
+		cache->health_timer = malloc(sizeof(*cache->health_timer));
+		if (!cache->health_timer) return kr_error(ENOMEM);
+		uv_loop_t *loop = uv_default_loop();
+		assert(loop);
+		int ret = uv_timer_init(loop, cache->health_timer);
+		if (ret) {
+			free(cache->health_timer);
+			cache->health_timer = NULL;
+			return kr_error(ret);
+		}
+		cache->health_timer->data = cache;
+	}
+	assert(cache->health_timer->data);
+	return kr_error(uv_timer_start(cache->health_timer, health_timer_cb, interval, interval));
+}
+
