@@ -474,11 +474,15 @@ static int process_authority(knot_pkt_t *pkt, struct kr_request *req)
 	return result;
 }
 
-static void finalize_answer(knot_pkt_t *pkt, struct kr_query *qry, struct kr_request *req)
+static int finalize_answer(knot_pkt_t *pkt, struct kr_request *req)
 {
 	/* Finalize header */
-	knot_pkt_t *answer = req->answer;
-	knot_wire_set_rcode(answer->wire, knot_wire_get_rcode(pkt->wire));
+	knot_pkt_t *answer = kr_request_ensure_answer(req);
+	if (answer) {
+		knot_wire_set_rcode(answer->wire, knot_wire_get_rcode(pkt->wire));
+		req->state = KR_STATE_DONE;
+	}
+	return req->state;
 }
 
 static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral, const knot_dname_t **cname_ret)
@@ -605,15 +609,17 @@ static int unroll_cname(knot_pkt_t *pkt, struct kr_request *req, bool referral,
 		}
 		/* The validator still can't handle multiple zones in one answer,
 		 * so we only follow if a single label is replaced.
-		 * TODO: this still isn't 100%, as the target might have a NS+DS,
-		 * possibly leading to a SERVFAIL for the in-bailiwick name. */
+		 * Forwarding appears to be even more sensitive to this.
+		 * TODO: iteration can probably handle the remaining cases,
+		 * but overall it would be better to have a smarter validator
+		 * (and thus save roundtrips).*/
 		const int pending_labels = knot_dname_labels(pending_cname, NULL);
 		if (pending_labels != cname_labels) {
 			cname = pending_cname;
 			break;
 		}
-		if (knot_dname_matched_labels(pending_cname, cname) !=
-		    (cname_labels - 1)) {
+		if (knot_dname_matched_labels(pending_cname, cname) != cname_labels - 1
+		    || query->flags.FORWARD) {
 			cname = pending_cname;
 			break;
 		}
@@ -677,8 +683,7 @@ static int process_final(knot_pkt_t *pkt, struct kr_request *req,
 			array->at[last_idx] = entry;
 			entry->to_wire = true;
 		}
-		finalize_answer(pkt, query, req);
-		return KR_STATE_DONE;
+		return finalize_answer(pkt, req);
 	}
 	return kr_ok();
 }
@@ -801,7 +806,7 @@ static int process_answer(knot_pkt_t *pkt, struct kr_request *req)
 		if (state != kr_ok()) {
 			return KR_STATE_FAIL;
 		}
-		finalize_answer(pkt, query, req);
+		return finalize_answer(pkt, req);
 	} else {
 		/* Answer for sub-query; DS, IP for NS etc.
 		 * It may contains NSEC \ NSEC3 records for
@@ -864,8 +869,7 @@ static int process_stub(knot_pkt_t *pkt, struct kr_request *req)
 		return KR_STATE_FAIL;
 	}
 
-	finalize_answer(pkt, query, req);
-	return KR_STATE_DONE;
+	return finalize_answer(pkt, req);
 }
 
 
@@ -902,7 +906,9 @@ static int begin(kr_layer_t *ctx)
 	if (qry->sclass != KNOT_CLASS_IN
 	    || (knot_rrtype_is_metatype(qry->stype)
 		    /* && qry->stype != KNOT_RRTYPE_ANY hmm ANY seems broken ATM */)) {
-		knot_wire_set_rcode(ctx->req->answer->wire, KNOT_RCODE_NOTIMPL);
+		knot_pkt_t *ans = kr_request_ensure_answer(ctx->req);
+		if (!ans) return ctx->req->state;
+		knot_wire_set_rcode(ans->wire, KNOT_RCODE_NOTIMPL);
 		return KR_STATE_FAIL;
 	}
 
@@ -986,36 +992,6 @@ static int resolve_badmsg(knot_pkt_t *pkt, struct kr_request *req, struct kr_que
 #endif
 }
 
-static int resolve_notimpl(knot_pkt_t *pkt, struct kr_request *req, struct kr_query *qry)
-{
-	if (qry->stype == KNOT_RRTYPE_RRSIG && qry->parent != NULL) {
-		/* RRSIG subquery have got NOTIMPL.
-		 * Possible scenario - same NS is autoritative for child and parent,
-		 * but child isn't signed.
-		 * We got delegation to parent,
-		 * then NS responded as NS for child zone.
-		 * Answer contained record been requested, but no RRSIGs,
-		 * Validator issued RRSIG query then. If qname is zone name,
-		 * we can get NOTIMPL. Ask for DS to find out security status.
-		 * TODO - maybe it would be better to do this in validator, when
-		 * RRSIG revalidation occurs.
-		 */
-		struct kr_rplan *rplan = &req->rplan;
-		struct kr_query *next = kr_rplan_push(rplan, qry->parent, qry->sname,
-						qry->sclass, KNOT_RRTYPE_DS);
-		if (!next) {
-			return KR_STATE_FAIL;
-		}
-		kr_zonecut_set(&next->zone_cut, qry->parent->zone_cut.name);
-		kr_zonecut_copy(&next->zone_cut, &qry->parent->zone_cut);
-		kr_zonecut_copy_trust(&next->zone_cut, &qry->parent->zone_cut);
-		next->flags.DNSSEC_WANT = true;
-		qry->flags.RESOLVED = true;
-		return KR_STATE_DONE;
-	}
-	return resolve_badmsg(pkt, req, qry);
-}
-
 /** Resolve input query or continue resolution with followups.
  *
  *  This roughly corresponds to RFC1034, 5.3.3 4a-d.
@@ -1091,6 +1067,8 @@ static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 	case KNOT_RCODE_NXDOMAIN:
 		break; /* OK */
 	case KNOT_RCODE_YXDOMAIN: /* Basically a successful answer; name just doesn't fit. */
+		if (!kr_request_ensure_answer(req))
+			return req->state;
 		knot_wire_set_rcode(req->answer->wire, KNOT_RCODE_YXDOMAIN);
 		break;
 	case KNOT_RCODE_REFUSED:
@@ -1101,11 +1079,9 @@ static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 		}
 		/* fall through */
 	case KNOT_RCODE_FORMERR:
-		VERBOSE_MSG("<= rcode: %s\n", rcode ? rcode->name : "??");
-		return resolve_badmsg(pkt, req, query);
 	case KNOT_RCODE_NOTIMPL:
 		VERBOSE_MSG("<= rcode: %s\n", rcode ? rcode->name : "??");
-		return resolve_notimpl(pkt, req, query);
+		return resolve_badmsg(pkt, req, query);
 	default:
 		VERBOSE_MSG("<= rcode: %s\n", rcode ? rcode->name : "??");
 		return resolve_error(pkt, req);
@@ -1134,7 +1110,8 @@ static int resolve(kr_layer_t *ctx, knot_pkt_t *pkt)
 	}
 
 rrarray_finalize:
-	/* Finish construction of libknot-format RRsets. */
+	/* Finish construction of libknot-format RRsets.
+	 * We do this even if dropping the answer, though it's probably useless. */
 	(void)0;
 	ranked_rr_array_t *selected[] = kr_request_selected(req);
 	for (knot_section_t i = KNOT_ANSWER; i <= KNOT_ADDITIONAL; ++i) {
