@@ -18,13 +18,13 @@
 #include "lib/layer/iterate.h"
 #include "lib/dnssec/ta.h"
 #include "lib/dnssec.h"
-#if defined(ENABLE_COOKIES)
+#if ENABLE_COOKIES
 #include "lib/cookies/control.h"
 #include "lib/cookies/helper.h"
 #include "lib/cookies/nonce.h"
 #else /* Define compatibility macros */
 #define KNOT_EDNS_OPTION_COOKIE 10
-#endif /* defined(ENABLE_COOKIES) */
+#endif /* ENABLE_COOKIES */
 
 #define VERBOSE_MSG(qry, ...) QRVERBOSE((qry), "resl",  __VA_ARGS__)
 
@@ -46,6 +46,17 @@ bool kr_rank_check(uint8_t rank)
 	}
 }
 
+bool kr_rank_test(uint8_t rank, uint8_t kr_flag)
+{
+	assert(kr_rank_check(rank) && kr_rank_check(kr_flag));
+	if (kr_flag == KR_RANK_AUTH) {
+		return rank & KR_RANK_AUTH;
+	}
+	assert(!(kr_flag & KR_RANK_AUTH));
+	/* The rest are exclusive values - exactly one has to be set. */
+	return (rank & ~KR_RANK_AUTH) == kr_flag;
+}
+
 /** @internal Set @a yielded to all RRs with matching @a qry_uid. */
 static void set_yield(ranked_rr_array_t *array, const uint32_t qry_uid, const bool yielded)
 {
@@ -409,16 +420,16 @@ static int edns_erase_and_reserve(knot_pkt_t *pkt)
 	return knot_pkt_reserve(pkt, len);
 }
 
-static int edns_create(knot_pkt_t *pkt, knot_pkt_t *template, struct kr_request *req)
+static int edns_create(knot_pkt_t *pkt, const struct kr_request *req)
 {
-	pkt->opt_rr = knot_rrset_copy(req->ctx->opt_rr, &pkt->mm);
+	pkt->opt_rr = knot_rrset_copy(req->ctx->upstream_opt_rr, &pkt->mm);
 	size_t wire_size = knot_edns_wire_size(pkt->opt_rr);
-#if defined(ENABLE_COOKIES)
+#if ENABLE_COOKIES
 	if (req->ctx->cookie_ctx.clnt.enabled ||
 	    req->ctx->cookie_ctx.srvr.enabled) {
 		wire_size += KR_COOKIE_OPT_MAX_LEN;
 	}
-#endif /* defined(ENABLE_COOKIES) */
+#endif /* ENABLE_COOKIES */
 	if (req->qsource.flags.tls) {
 		if (req->ctx->tls_padding == -1)
 			/* FIXME: we do not know how to reserve space for the
@@ -431,26 +442,6 @@ static int edns_create(knot_pkt_t *pkt, knot_pkt_t *template, struct kr_request
 	return knot_pkt_reserve(pkt, wire_size);
 }
 
-static int answer_prepare(struct kr_request *req, knot_pkt_t *query)
-{
-	knot_pkt_t *answer = req->answer;
-	if (knot_pkt_init_response(answer, query) != 0) {
-		return kr_error(ENOMEM); /* Failed to initialize answer */
-	}
-	/* Handle EDNS in the query */
-	if (knot_pkt_has_edns(query)) {
-		answer->opt_rr = knot_rrset_copy(req->ctx->opt_rr, &answer->mm);
-		if (answer->opt_rr == NULL){
-			return kr_error(ENOMEM);
-		}
-		/* Set DO bit if set (DNSSEC requested). */
-		if (knot_pkt_has_dnssec(query)) {
-			knot_edns_set_do(answer->opt_rr);
-		}
-	}
-	return kr_ok();
-}
-
 /**
  * @param all_secure optionally &&-combine security of written RRs into its value.
  *		     (i.e. if you pass a pointer to false, it will always remain)
@@ -542,6 +533,9 @@ static int answer_padding(struct kr_request *request)
 static void answer_fail(struct kr_request *request)
 {
 	/* Note: OPT in SERVFAIL response is still useful for cookies/additional info. */
+	if (VERBOSE_STATUS || kr_log_rtrace_enabled(request))
+		kr_log_req(request, 0, 0, "resl",
+			"request failed, answering with empty SERVFAIL\n");
 	knot_pkt_t *answer = request->answer;
 	knot_rrset_t *opt_rr = answer->opt_rr; /* it gets NULLed below */
 	int ret = kr_pkt_clear_payload(answer);
@@ -573,6 +567,7 @@ static void answer_finalize(struct kr_request *request)
 {
 	struct kr_rplan *rplan = &request->rplan;
 	knot_pkt_t *answer = request->answer;
+	const uint8_t *q_wire = request->qsource.packet->wire;
 
 	if (answer->rrset_count != 0) {
 		/* Non-standard: we assume the answer had been constructed.
@@ -611,7 +606,7 @@ static void answer_finalize(struct kr_request *request)
 	/* TODO: clean this up in !660 or followup, and it isn't foolproof anyway. */
 	if (last->flags.DNSSEC_BOGUS
 	    || (rplan->pending.len > 0 && array_tail(rplan->pending)->flags.DNSSEC_BOGUS)) {
-		if (!knot_wire_get_cd(request->qsource.packet->wire)) {
+		if (!knot_wire_get_cd(q_wire)) {
 			answer_fail(request);
 			return;
 		}
@@ -626,7 +621,7 @@ static void answer_finalize(struct kr_request *request)
 		secure = false; /* don't trust forwarding for now */
 	}
 	if (last && (last->flags.DNSSEC_OPTOUT)) {
-		VERBOSE_MSG(NULL, "AD: opt-out\n");
+		VERBOSE_MSG(last, "insecure because of opt-out\n");
 		secure = false; /* the last answer is insecure due to opt-out */
 	}
 
@@ -676,9 +671,10 @@ static void answer_finalize(struct kr_request *request)
 	VERBOSE_MSG(last, "AD: request%s classified as SECURE\n", secure ? "" : " NOT");
 	request->rank = secure ? KR_RANK_SECURE : KR_RANK_INITIAL;
 
-	/* Clear AD if not secure.  ATM answer has AD=1 if requested secured answer. */
-	if (!secure) {
-		knot_wire_clear_ad(answer->wire);
+	/* Set AD if secure and AD bit "was requested". */
+	if (secure && !knot_wire_get_cd(q_wire)
+	    && (knot_pkt_has_dnssec(answer) || knot_wire_get_ad(q_wire))) {
+		knot_wire_set_ad(answer->wire);
 	}
 }
 
@@ -690,7 +686,7 @@ static int query_finalize(struct kr_request *request, struct kr_query *qry, knot
 	/* Remove any EDNS records from any previous iteration. */
 	int ret = edns_erase_and_reserve(pkt);
 	if (ret) return ret;
-	ret = edns_create(pkt, request->answer, request);
+	ret = edns_create(pkt, request);
 	if (ret) return ret;
 	if (qry->flags.STUB) {
 		/* Stub resolution (ask for +rd and +do) */
@@ -712,11 +708,11 @@ static int query_finalize(struct kr_request *request, struct kr_query *qry, knot
 	return kr_ok();
 }
 
-int kr_resolve_begin(struct kr_request *request, struct kr_context *ctx, knot_pkt_t *answer)
+int kr_resolve_begin(struct kr_request *request, struct kr_context *ctx)
 {
 	/* Initialize request */
 	request->ctx = ctx;
-	request->answer = answer;
+	request->answer = NULL;
 	request->options = ctx->options;
 	request->state = KR_STATE_CONSUME;
 	request->current_query = NULL;
@@ -762,25 +758,11 @@ static int resolve_query(struct kr_request *request, const knot_pkt_t *packet)
 		qry->flags.AWAIT_CUT = true;
 		/* Want DNSSEC if it's posible to secure this name (e.g. is covered by any TA) */
 		if ((knot_wire_get_ad(packet->wire) || knot_pkt_has_dnssec(packet)) &&
-		    kr_ta_covers_qry(request->ctx, qname, qtype)) {
+		    kr_ta_covers_qry(request->ctx, qry->sname, qtype)) {
 			qry->flags.DNSSEC_WANT = true;
 		}
 	}
 
-	/* Initialize answer packet */
-	knot_pkt_t *answer = request->answer;
-	knot_wire_set_qr(answer->wire);
-	knot_wire_clear_aa(answer->wire);
-	knot_wire_set_ra(answer->wire);
-	knot_wire_set_rcode(answer->wire, KNOT_RCODE_NOERROR);
-
-	assert(request->qsource.packet);
-	if (knot_wire_get_cd(request->qsource.packet->wire)) {
-		knot_wire_set_cd(answer->wire);
-	} else if (qry->flags.DNSSEC_WANT) {
-		knot_wire_set_ad(answer->wire);
-	}
-
 	/* Expect answer, pop if satisfied immediately */
 	ITERATE_LAYERS(request, qry, begin);
 	if ((request->state & KR_STATE_DONE) != 0) {
@@ -794,6 +776,69 @@ static int resolve_query(struct kr_request *request, const knot_pkt_t *packet)
 	return request->state;
 }
 
+knot_pkt_t * kr_request_ensure_answer(struct kr_request *request)
+{
+	if (request->answer)
+		return request->answer;
+
+	const knot_pkt_t *qs_pkt = request->qsource.packet;
+	assert(qs_pkt);
+	// Find answer_max: limit on DNS wire length.
+	uint16_t answer_max;
+	const struct kr_request_qsource_flags *qs_flags = &request->qsource.flags;
+	assert((qs_flags->tls || qs_flags->http) ? qs_flags->tcp : true);
+	if (!request->qsource.addr || qs_flags->tcp) {
+		// not on UDP
+		answer_max = KNOT_WIRE_MAX_PKTSIZE;
+	} else if (knot_pkt_has_edns(qs_pkt)) {
+		// UDP with EDNS
+		answer_max = MIN(knot_edns_get_payload(qs_pkt->opt_rr),
+				 knot_edns_get_payload(request->ctx->downstream_opt_rr));
+		answer_max = MAX(answer_max, KNOT_WIRE_MIN_PKTSIZE);
+	} else {
+		// UDP without EDNS
+		answer_max = KNOT_WIRE_MIN_PKTSIZE;
+	}
+
+	// Allocate the packet.
+	uint8_t *wire = NULL;
+	if (request->alloc_wire_cb) {
+		wire = request->alloc_wire_cb(request, &answer_max);
+		if (!wire)
+			goto enomem;
+	}
+	knot_pkt_t *answer = request->answer =
+		knot_pkt_new(wire, answer_max, &request->pool);
+	if (!answer || knot_pkt_init_response(answer, qs_pkt) != 0) {
+		assert(!answer); // otherwise we messed something up
+		goto enomem;
+	}
+	if (!wire)
+		wire = answer->wire;
+
+	// Much was done by knot_pkt_init_response()
+	knot_wire_set_ra(wire);
+	knot_wire_set_rcode(wire, KNOT_RCODE_NOERROR);
+	if (knot_wire_get_cd(qs_pkt->wire)) {
+		knot_wire_set_cd(wire);
+	}
+
+	// Prepare EDNS if required.
+	if (knot_pkt_has_edns(qs_pkt)) {
+		answer->opt_rr = knot_rrset_copy(request->ctx->downstream_opt_rr,
+						 &answer->mm);
+		if (!answer->opt_rr)
+			goto enomem; // answer is on mempool, so "leak" is OK
+		if (knot_pkt_has_dnssec(qs_pkt))
+			knot_edns_set_do(answer->opt_rr);
+	}
+
+	return request->answer;
+enomem:
+	request->state = KR_STATE_FAIL; // TODO: really combine with another flag?
+	return request->answer = NULL;
+}
+
 KR_PURE static bool kr_inaddr_equal(const struct sockaddr *a, const struct sockaddr *b)
 {
 	const int a_len = kr_inaddr_len(a);
@@ -890,9 +935,6 @@ int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, k
 
 	/* Empty resolution plan, push packet as the new query */
 	if (packet && kr_rplan_empty(rplan)) {
-		if (answer_prepare(request, packet) != 0) {
-			return KR_STATE_FAIL;
-		}
 		return resolve_query(request, packet);
 	}
 
@@ -947,7 +989,7 @@ int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, k
 					kr_log_req(request, 0, 2, "resl",
 						"=> too many failures in a row, "
 						"bail out (mitigation for NXNSAttack "
-						"CVE-2020-12667)");
+						"CVE-2020-12667)\n");
 				}
 				return KR_STATE_FAIL;
 			}
@@ -1200,6 +1242,7 @@ static int trust_chain_check(struct kr_request *request, struct kr_query *qry)
 	if (qry->flags.DNSSEC_NODS) {
 		/* This is the next query iteration with minimized qname.
 		 * At previous iteration DS non-existance has been proven */
+		VERBOSE_MSG(qry, "<= DS doesn't exist, going insecure\n");
 		qry->flags.DNSSEC_NODS = false;
 		qry->flags.DNSSEC_WANT = false;
 		qry->flags.DNSSEC_INSECURE = true;
@@ -1493,7 +1536,7 @@ ns_election:
 	return request->state;
 }
 
-#if defined(ENABLE_COOKIES)
+#if ENABLE_COOKIES
 /** Update DNS cookie data in packet. */
 static bool outbound_request_update_cookies(struct kr_request *req,
                                             const struct sockaddr *src,
@@ -1523,7 +1566,7 @@ static bool outbound_request_update_cookies(struct kr_request *req,
 
 	return true;
 }
-#endif /* defined(ENABLE_COOKIES) */
+#endif /* ENABLE_COOKIES */
 
 int kr_resolve_checkout(struct kr_request *request, const struct sockaddr *src,
                         struct sockaddr *dst, int type, knot_pkt_t *packet)
@@ -1542,7 +1585,7 @@ int kr_resolve_checkout(struct kr_request *request, const struct sockaddr *src,
 	}
 	struct kr_query *qry = array_tail(rplan->pending);
 
-#if defined(ENABLE_COOKIES)
+#if ENABLE_COOKIES
 	/* Update DNS cookies in request. */
 	if (type == SOCK_DGRAM) { /* @todo: Add cookies also over TCP? */
 		/*
@@ -1554,7 +1597,7 @@ int kr_resolve_checkout(struct kr_request *request, const struct sockaddr *src,
 			return kr_error(EINVAL);
 		}
 	}
-#endif /* defined(ENABLE_COOKIES) */
+#endif /* ENABLE_COOKIES */
 
 	int ret = query_finalize(request, qry, packet);
 	if (ret != 0) {
@@ -1618,20 +1661,23 @@ int kr_resolve_checkout(struct kr_request *request, const struct sockaddr *src,
 int kr_resolve_finish(struct kr_request *request, int state)
 {
 	request->state = state;
-	/* Finalize answer and construct wire-buffer. */
-	ITERATE_LAYERS(request, NULL, answer_finalize);
-	answer_finalize(request);
+	/* Finalize answer and construct whole wire-format (unless dropping). */
+	knot_pkt_t *answer = kr_request_ensure_answer(request);
+	if (answer) {
+		ITERATE_LAYERS(request, NULL, answer_finalize);
+		answer_finalize(request);
 
-	/* Defensive style, in case someone has forgotten.
-	 * Beware: non-empty answers do make sense even with SERVFAIL case, etc. */
-	if (request->state != KR_STATE_DONE) {
-		uint8_t *wire = request->answer->wire;
-		switch (knot_wire_get_rcode(wire)) {
-		case KNOT_RCODE_NOERROR:
-		case KNOT_RCODE_NXDOMAIN:
-			knot_wire_clear_ad(wire);
-			knot_wire_clear_aa(wire);
-			knot_wire_set_rcode(wire, KNOT_RCODE_SERVFAIL);
+		/* Defensive style, in case someone has forgotten.
+		 * Beware: non-empty answers do make sense even with SERVFAIL case, etc. */
+		if (request->state != KR_STATE_DONE) {
+			uint8_t *wire = answer->wire;
+			switch (knot_wire_get_rcode(wire)) {
+			case KNOT_RCODE_NOERROR:
+			case KNOT_RCODE_NXDOMAIN:
+				knot_wire_clear_ad(wire);
+				knot_wire_clear_aa(wire);
+				knot_wire_set_rcode(wire, KNOT_RCODE_SERVFAIL);
+			}
 		}
 	}
 
@@ -1640,7 +1686,7 @@ int kr_resolve_finish(struct kr_request *request, int state)
 #ifndef NOVERBOSELOG
 	struct kr_rplan *rplan = &request->rplan;
 	struct kr_query *last = kr_rplan_last(rplan);
-	VERBOSE_MSG(last, "finished: %d, queries: %zu, mempool: %zu B\n",
+	VERBOSE_MSG(last, "finished in state: %d, queries: %zu, mempool: %zu B\n",
 	          request->state, rplan->resolved.len, (size_t) mp_total_size(request->pool.ctx));
 #endif
 
