@@ -36,7 +36,7 @@
  * };
  *
  * // Setup and provide input query
- * int state = kr_resolve_begin(&req, ctx, final_answer);
+ * int state = kr_resolve_begin(&req, ctx);
  * state = kr_resolve_consume(&req, query);
  *
  * // Generate answer
@@ -61,6 +61,20 @@
  */
 
 
+struct kr_request;
+/** Allocate buffer for answer's wire (*maxlen may get lowered).
+ *
+ * Motivation: XDP wire allocation is an overlap of library and daemon:
+ *  - it needs to be called from the library
+ *  - it needs to rely on some daemon's internals
+ *  - the library (currently) isn't allowed to directly use symbols from daemon
+ *    (contrary to modules), e.g. some of our lib-using tests run without daemon
+ *
+ * Note: after we obtain the wire, we're obliged to send it out.
+ * (So far there's no use case to allow cancelling at that point.)
+ */
+typedef uint8_t * (*alloc_wire_f)(struct kr_request *req, uint16_t *maxlen);
+
 /**
  * RRset rank - for cache and ranked_rr_*.
  *
@@ -90,7 +104,7 @@ enum kr_rank {
 	KR_RANK_INDET = 4,   /**< Unable to determine whether it should be secure. */
 	KR_RANK_BOGUS,       /**< Ought to be secure but isn't. */
 	KR_RANK_MISMATCH,
-	KR_RANK_MISSING,     /**< Unable to obtain a good signature. */
+	KR_RANK_MISSING,     /**< No RRSIG found for that owner+type combination. */
 
 	/** Proven to be insecure, i.e. we have a chain of trust from TAs
 	 * that cryptographically denies the possibility of existence
@@ -110,16 +124,7 @@ enum kr_rank {
 bool kr_rank_check(uint8_t rank) KR_PURE;
 
 /** Test the presence of any flag/state in a rank, i.e. including KR_RANK_AUTH. */
-static inline bool kr_rank_test(uint8_t rank, uint8_t kr_flag)
-{
-	assert(kr_rank_check(rank) && kr_rank_check(kr_flag));
-	if (kr_flag == KR_RANK_AUTH) {
-		return rank & KR_RANK_AUTH;
-	}
-	assert(!(kr_flag & KR_RANK_AUTH));
-	/* The rest are exclusive values - exactly one has to be set. */
-	return (rank & ~KR_RANK_AUTH) == kr_flag;
-}
+bool kr_rank_test(uint8_t rank, uint8_t kr_flag) KR_PURE KR_EXPORT;
 
 /** Set the rank state. The _AUTH flag is kept as it was. */
 static inline void kr_rank_set(uint8_t *rank, uint8_t kr_flag)
@@ -149,7 +154,8 @@ struct kr_context
 	/** Default EDNS towards *both* clients and upstream.
 	 * LATER: consider splitting the two, e.g. allow separately
 	 * configured limits for UDP packet size (say, LAN is under control). */
-	knot_rrset_t *opt_rr;
+	knot_rrset_t *downstream_opt_rr;
+	knot_rrset_t *upstream_opt_rr;
 
 	map_t trust_anchors;
 	map_t negative_anchors;
@@ -173,6 +179,7 @@ struct kr_request_qsource_flags {
 	bool tcp:1; /**< true if the request is not on UDP; only meaningful if (dst_addr). */
 	bool tls:1; /**< true if the request is encrypted; only meaningful if (dst_addr). */
 	bool http:1; /**< true if the request is on HTTP; only meaningful if (dst_addr). */
+	bool xdp:1; /**< true if the request is on AF_XDP; only meaningful if (dst_addr). */
 };
 
 /**
@@ -187,7 +194,7 @@ struct kr_request_qsource_flags {
  */
 struct kr_request {
 	struct kr_context *ctx;
-	knot_pkt_t *answer;
+	knot_pkt_t *answer; /**< See kr_request_ensure_answer() */
 	struct kr_query *current_query;    /**< Current evaluated query. */
 	struct {
 		/** Address that originated the request. NULL for internal origin. */
@@ -199,6 +206,7 @@ struct kr_request {
 		const knot_pkt_t *packet;
 		struct kr_request_qsource_flags flags; /**< See definition above. */
 		size_t size; /**< query packet size */
+		int32_t stream_id; /**< HTTP/2 stream ID for DoH requests */
 	} qsource;
 	struct {
 		unsigned rtt;                  /**< Current upstream RTT */
@@ -226,9 +234,10 @@ struct kr_request {
 	trace_callback_f trace_finish; /**< Request finish tracepoint */
 	int vars_ref; /**< Reference to per-request variable table. LUA_NOREF if not set. */
 	knot_mm_t pool;
-	unsigned int uid; /** for logging purposes only */
+	unsigned int uid; /**< for logging purposes only */
 	unsigned int count_no_nsaddr;
 	unsigned int count_fail_row;
+	alloc_wire_f alloc_wire_cb; /**< CB to allocate answer wire (can be NULL). */
 };
 
 /** Initializer for an array of *_selected. */
@@ -241,22 +250,31 @@ struct kr_request {
 /**
  * Begin name resolution.
  *
- * @note Expects a request to have an initialized mempool, the "answer" packet will
- *       be kept during the resolution and will contain the final answer at the end.
+ * @note Expects a request to have an initialized mempool.
  *
  * @param request request state with initialized mempool
  * @param ctx     resolution context
- * @param answer  allocated packet for final answer
  * @return        CONSUME (expecting query)
  */
 KR_EXPORT
-int kr_resolve_begin(struct kr_request *request, struct kr_context *ctx, knot_pkt_t *answer);
+int kr_resolve_begin(struct kr_request *request, struct kr_context *ctx);
+
+/**
+ * Ensure that request->answer is usable, and return it (for convenience).
+ *
+ * It may return NULL, in which case it marks ->state with _FAIL and no answer will be sent.
+ * Only use this when it's guaranteed that there will be no delay before sending it.
+ * You don't need to call this in places where "resolver knows" that there will be no delay,
+ * but even there you need to check if the ->answer is NULL (unless you check for _FAIL anyway).
+ */
+KR_EXPORT
+knot_pkt_t * kr_request_ensure_answer(struct kr_request *request);
 
 /**
  * Consume input packet (may be either first query or answer to query originated from kr_resolve_produce())
  *
  * @note If the I/O fails, provide an empty or NULL packet, this will make iterator recognize nameserver failure.
- * 
+ *
  * @param  request request state (awaiting input)
  * @param  src     [in] packet source address
  * @param  packet  [in] input packet
@@ -271,7 +289,7 @@ int kr_resolve_consume(struct kr_request *request, const struct sockaddr *src, k
  * If the CONSUME is returned then dst, type and packet will be filled with
  * appropriate values and caller is responsible to send them and receive answer.
  * If it returns any other state, then content of the variables is undefined.
- * 
+ *
  * @param  request request state (in PRODUCE state)
  * @param  dst     [out] possible address of the next nameserver
  * @param  type    [out] possible used socket type (SOCK_STREAM, SOCK_DGRAM)
