@@ -210,6 +210,31 @@ char* kr_strcatdup(unsigned n, ...)
 	return result;
 }
 
+char * kr_absolutize_path(const char *dirname, const char *fname)
+{
+	assert(dirname && fname);
+	char *result;
+	int aret;
+	if (dirname[0] == '/') { // absolute path is easier
+		aret = asprintf(&result, "%s/%s", dirname, fname);
+	} else { // relative path, but don't resolve symlinks
+		char buf[PATH_MAX];
+		const char *cwd = getcwd(buf, sizeof(buf));
+		if (!cwd)
+			return NULL; // errno has been set already
+		if (strcmp(dirname, ".") == 0) {
+			// get rid of one common case of extraneous "./"
+			aret = asprintf(&result, "%s/%s", cwd, fname);
+		} else {
+			aret = asprintf(&result, "%s/%s/%s", cwd, dirname, fname);
+		}
+	}
+	if (aret > 0)
+		return result;
+	errno = -aret;
+	return NULL;
+}
+
 int kr_memreserve(void *baton, void **mem, size_t elm_size, size_t want, size_t *have)
 {
     if (*have >= want) {
@@ -1056,8 +1081,7 @@ char *kr_pkt_text(const knot_pkt_t *pkt)
 
 	for (knot_section_t i = KNOT_ANSWER; i <= KNOT_ADDITIONAL; ++i) {
 		const knot_pktsection_t *sec = knot_pkt_section(pkt, i);
-		if (sec->count == 0 || knot_pkt_rr(sec, 0)->type == KNOT_RRTYPE_OPT) {
-			/* OPT RRs are _supposed_ to be the last ^^, if they appear */
+		if (sec->count == 0) {
 			continue;
 		}
 
@@ -1266,3 +1290,8 @@ long long kr_fssize(const char *path)
 	return buf.f_frsize * buf.f_blocks;
 }
 
+const char * kr_dirent_name(const struct dirent *de)
+{
+	return de ? de->d_name : NULL;
+}
+
