@@ -1,17 +1,5 @@
 /*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program.  If not, see <https://www.gnu.org/licenses/>.
+ *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #include "lib/zonecut.h"
@@ -400,7 +388,7 @@ static int fetch_ns(struct kr_context *ctx, struct kr_zonecut *cut,
 		return kr_error(ESTALE);
 	}
 	/* Materialize the rdataset temporarily, for simplicity. */
-	knot_rdataset_t ns_rds = { 0, NULL };
+	knot_rdataset_t ns_rds = { 0 };
 	ret = kr_cache_materialize(&ns_rds, &peek, cut->pool);
 	if (ret < 0) {
 		return ret;
@@ -591,7 +579,7 @@ int kr_zonecut_find_cached(struct kr_context *ctx, struct kr_zonecut *cut,
 		}
 	}
 
-	kr_cache_sync(&ctx->cache);
+	kr_cache_commit(&ctx->cache);
 	mm_free(cut->pool, qname);
 	return ret;
 }
