@@ -33,7 +33,7 @@ struct dnstap_data {
 
 /*
  * dt_pack packs the dnstap message for transport
- * https://gitlab.labs.nic.cz/knot/knot-dns/blob/master/src/contrib/dnstap/dnstap.c#L24
+ * https://gitlab.nic.cz/knot/knot-dns/blob/master/src/contrib/dnstap/dnstap.c#L24
  * */
 uint8_t* dt_pack(const Dnstap__Dnstap *d, uint8_t **buf, size_t *sz)
 {
@@ -54,7 +54,7 @@ uint8_t* dt_pack(const Dnstap__Dnstap *d, uint8_t **buf, size_t *sz)
 }
 
 /* set_address fills in address detail in dnstap_message
- * https://gitlab.labs.nic.cz/knot/knot-dns/blob/master/src/contrib/dnstap/message.c#L28
+ * https://gitlab.nic.cz/knot/knot-dns/blob/master/src/contrib/dnstap/message.c#L28
  */
 static void set_address(const struct sockaddr *sockaddr,
 		ProtobufCBinaryData   *addr,
@@ -136,9 +136,11 @@ static int dnstap_log(kr_layer_t *ctx) {
 
 	if (dnstap_dt->log_resp_pkt) {
 		const knot_pkt_t *rpkt = req->answer;
-		m.response_message.len = rpkt->size;
-		m.response_message.data = (uint8_t *)rpkt->wire;
-		m.has_response_message = true;
+		m.has_response_message = rpkt != NULL;
+		if (rpkt != NULL) {
+			m.response_message.len = rpkt->size;
+			m.response_message.data = (uint8_t *)rpkt->wire;
+		}
 	}
 
 	/* set query time to the timestamp of the first kr_query
@@ -232,7 +234,7 @@ int dnstap_deinit(struct kr_module *module) {
 }
 
 /* dnstap_unix_writer returns a unix fstream writer
- * https://gitlab.labs.nic.cz/knot/knot-dns/blob/master/src/knot/modules/dnstap.c#L159
+ * https://gitlab.nic.cz/knot/knot-dns/blob/master/src/knot/modules/dnstap.c#L159
  */
 static struct fstrm_writer* dnstap_unix_writer(const char *path) {
 
