@@ -159,7 +159,7 @@ class Entry:
         # MATCH
         self.match_fields = self.process_match()
 
-        # FLAGS
+        # FLAGS (old alias REPLY)
         self.process_reply_line()
 
         # ADJUST
@@ -204,12 +204,12 @@ class Entry:
 
     def process_reply_line(self):
         """Extracts flags, rcode and opcode from given node and adjust dns message accordingly"""
-        self.fields = [f.value for f in self.node.match("/reply")]
-        if 'DO' in self.fields:
+        fields = [f.value for f in self.node.match("/reply")]
+        if 'DO' in fields:
             self.message.want_dnssec(True)
-        opcode = self.get_opcode(fields=self.fields)
-        rcode = self.get_rcode(fields=self.fields)
-        self.message.flags = self.get_flags(fields=self.fields)
+        opcode = self.get_opcode(fields)
+        rcode = self.get_rcode(fields)
+        self.message.flags = self.get_flags(fields)
         if rcode is not None:
             self.message.set_rcode(rcode)
         if opcode is not None:
@@ -248,7 +248,7 @@ class Entry:
                     rd = record['/data'].value.split()
                     if rd:
                         if rdtype == dns.rdatatype.DS:
-                            rd[1] = str(dns.dnssec.algorithm_from_text(rd[1]))
+                            rd[1] = '{}'.format(dns.dnssec.algorithm_from_text(rd[1]))
                         rd = dns.rdata.from_text(rr.rdclass, rr.rdtype, ' '.join(
                             rd), origin=dns.name.from_text(self.origin), relativize=False)
                     rr.add(rd)
@@ -348,7 +348,7 @@ class Entry:
             except pydnstest.matchpart.DataMismatch as ex:
                 errstr = '%s in the response:\n%s' % (str(ex), msg.to_text())
                 # TODO: cisla radku
-                raise ValueError("%s, \"%s\": %s" % (self.node.span, code, errstr))
+                raise ValueError("%s, \"%s\": %s" % (self.node.span, code, errstr)) from None
 
     def cmp_raw(self, raw_value):
         assert self.raw_data is not None
@@ -584,9 +584,10 @@ class Step:
                                                   ctx.client[choice][1],
                                                   tcp,
                                                   src_address=src_address)
-        pydnstest.mock_client.send_query(sock, data_to_wire)
-        if self.data[0].raw_data is None:
-            answer = pydnstest.mock_client.get_answer(sock)
+        with sock:
+            pydnstest.mock_client.send_query(sock, data_to_wire)
+            if self.data[0].raw_data is None:
+                answer = pydnstest.mock_client.get_answer(sock)
 
         # Track RTT
         rtt = (datetime.now() - tstart).total_seconds() * 1000
@@ -696,15 +697,16 @@ class Scenario:
                             j].id == step.next_if_fail]
                         if not next_steps:
                             raise ValueError('step %d: wrong NEXT value "%d"' %
-                                             (step.id, step.next_if_fail))
+                                             (step.id, step.next_if_fail)) from ex
                         next_step = next_steps[0]
                         if next_step < len(self.steps):
                             i = next_step
                         else:
                             raise ValueError('step %d: Can''t branch to NEXT value "%d"' %
-                                             (step.id, step.next_if_fail))
+                                             (step.id, step.next_if_fail)) from ex
                     continue
-                raise ValueError('%s step %d %s' % (self.file, step.id, str(ex)))
+                ex_details = ex if self.log.isEnabledFor(logging.DEBUG) else None
+                raise ValueError('%s step %d %s' % (self.file, step.id, str(ex))) from ex_details
             i += 1
 
         for r in self.ranges:
@@ -756,6 +758,7 @@ def parse_config(scn_cfg, qmin, installdir):  # FIXME: pylint: disable=too-many-
     negative_ta_list = []
     stub_addr = None
     override_timestamp = None
+    forward_addr = None
 
     features = {}
     feature_list_delimiter = ';'
@@ -808,7 +811,8 @@ def parse_config(scn_cfg, qmin, installdir):  # FIXME: pylint: disable=too-many-
                         f_value = ""
                     features[f_key] = f_value
             except KeyError as ex:
-                raise KeyError("can't parse features (%s) in config section (%s)" % (v, str(ex)))
+                raise KeyError("can't parse features (%s) in config section (%s)"
+                               % (v, str(ex))) from ex
         elif k == 'feature-list':
             try:
                 f_key, f_value = [x.strip() for x in v.split(feature_pair_delimiter, 1)]
@@ -818,9 +822,11 @@ def parse_config(scn_cfg, qmin, installdir):  # FIXME: pylint: disable=too-many-
                 features[f_key].append(f_value)
             except KeyError as ex:
                 raise KeyError("can't parse feature-list (%s) in config section (%s)"
-                               % (v, str(ex)))
+                               % (v, str(ex))) from ex
         elif k == 'force-ipv6' and v.upper() == 'TRUE':
             sockfamily = socket.AF_INET6
+        elif k == 'forward-addr':  # currently forwards everything
+            forward_addr = v.strip('"\'')
         else:
             raise NotImplementedError('unsupported CONFIG key "%s"' % k)
 
@@ -832,7 +838,8 @@ def parse_config(scn_cfg, qmin, installdir):  # FIXME: pylint: disable=too-many-
         "INSTALL_DIR": installdir,
         "QMIN": str(qmin).lower(),
         "TRUST_ANCHORS": trust_anchor_list,
-        "TRUST_ANCHOR_FILES": trust_anchor_files
+        "TRUST_ANCHOR_FILES": trust_anchor_files,
+        "FORWARD_ADDR": forward_addr,
     }
     if stub_addr:
         ctx['ROOT_ADDR'] = stub_addr
